# Source: https://github.com/python/pyperformance
# License: MIT

# The Computer Language Benchmarks Game
# http://benchmarksgame.alioth.debian.org/
# Contributed by Sokolov Yura, modified by Tupteq.


def fannkuch(n):
    count = list(range(1, n + 1))
    max_flips = 0
    m = n - 1
    r = n
    check = 0
    perm1 = list(range(n))
    perm = list(range(n))
    perm1_ins = perm1.insert
    perm1_pop = perm1.pop

    while 1:
        if check < 30:
            check += 1

        while r != 1:
            count[r - 1] = r
            r -= 1

        if perm1[0] != 0 and perm1[m] != m:
            perm = perm1[:]
            flips_count = 0
            k = perm[0]
            while k:
                perm[: k + 1] = perm[k::-1]
                flips_count += 1
                k = perm[0]

            if flips_count > max_flips:
                max_flips = flips_count

        while r != n:
            perm1_ins(r, perm1_pop(0))
            count[r] -= 1
            if count[r] > 0:
                break
            r += 1
        else:
            return max_flips


###########################################################################
# Benchmark interface

bm_params = {
    (32, 10): (3,),
    (50, 10): (5,),
    (100, 10): (6,),
    (500, 10): (7,),
    (1000, 10): (8,),
    (5000, 10): (9,),
}


def bm_setup(params):
    state = None

    def run():
        nonlocal state
        state = fannkuch(params[0])

    def result():
        return params[0], state

    return run, result



try:
    import io
    import errno
    import websocket
except ImportError:
    print("SKIP")
    raise SystemExit


# put raw data in the stream and do a websocket read
def ws_read(msg, sz):
    ws = websocket.websocket(io.BytesIO(msg))
    return ws.read(sz)


# do a websocket write and then return the raw data from the stream
def ws_write(msg, sz):
    s = io.BytesIO()
    ws = websocket.websocket(s)
    ws.write(msg)
    s.seek(0)
    return s.read(sz)


# basic frame
print(ws_read(b"\x81\x04ping", 4))
print(ws_read(b"\x80\x04ping", 4))  # FRAME_CONT
print(ws_write(b"pong", 6))

# split frames are not supported
# print(ws_read(b"\x01\x04ping", 4))

# extended payloads
print(ws_read(b"\x81~\x00\x80" + b"ping" * 32, 128))
print(ws_write(b"pong" * 32, 132))

# mask (returned data will be 'mask' ^ 'mask')
print(ws_read(b"\x81\x84maskmask", 4))

# close control frame
s = io.BytesIO(b"\x88\x00")  # FRAME_CLOSE
ws = websocket.websocket(s)
print(ws.read(1))
s.seek(2)
print(s.read(4))

# misc control frames
print(ws_read(b"\x89\x00\x81\x04ping", 4))  # FRAME_PING
print(ws_read(b"\x8a\x00\x81\x04pong", 4))  # FRAME_PONG

# close method
ws = websocket.websocket(io.BytesIO())
ws.close()

# ioctl
ws = websocket.websocket(io.BytesIO())
print(ws.ioctl(8))  # GET_DATA_OPTS
print(ws.ioctl(9, 2))  # SET_DATA_OPTS
print(ws.ioctl(9))
try:
    ws.ioctl(-1)
except OSError as e:
    print("ioctl: EINVAL:", e.errno == errno.EINVAL)



"""
The MIT License (MIT)

Copyright (c) 2023 Arduino SA
Copyright (c) 2018 KPN (Jan Bogaerts)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from senml import *

import time


class Coordinates(SenmlRecord):
    def __init__(self, name, **kwargs):
        """overriding the init function so we can initiate the 3 senml records that will represent lat,lon, alt"""
        self._lat = SenmlRecord(
            "lattitude", unit=SenmlUnits.SENML_UNIT_DEGREES_LATITUDE
        )  # create these befor calling base constructor so that all can be init correctly from constructor
        self._lon = SenmlRecord("longitude", unit=SenmlUnits.SENML_UNIT_DEGREES_LONGITUDE)
        self._alt = SenmlRecord("altitude", unit=SenmlUnits.SENML_UNIT_METER)
        super(Coordinates, self).__init__(
            name, **kwargs
        )  # need to call base init, to make certain all is ok.

    def _check_value_type(self, value):
        """overriding the check on value type to make certain that only an array with 3 values is assigned: lat,lon/alt"""
        if value is not None:
            if not isinstance(value, list):
                raise Exception("invalid data type: array with 3 elements expected lat, lon, alt")

    def _build_rec_dict(self, naming_map, appendTo):
        """
        override the rendering of the senml data objects. These will be converted to json or cbor
        :param naming_map: {dictionary} a map that determines the field names, these are different for json vs cbor
        :param appendTo: {list} the result list
        :return: None
        """
        self._lat._build_rec_dict(naming_map, appendTo)
        self._lon._build_rec_dict(naming_map, appendTo)
        self._alt._build_rec_dict(naming_map, appendTo)

    @SenmlRecord.value.setter
    def value(self, value):
        """set the current value.
        this is overridden so we can pass on the values to the internal objects. It's also stored in the parent
        so that a 'get-value' still returns the array.
        """
        self._value = (
            value  # micropython doesn't support calling setter of parent property, do it manually
        )
        if value:
            self._lat.value = value[0]
            self._lon.value = value[1]
            self._alt.value = value[2]
        else:
            self._lat.value = None
            self._lon.value = None
            self._alt.value = None

    @SenmlRecord.time.setter
    def time(self, value):
        """set the time stamp.
        this is overridden so we can pass on the values to the internal objects.
        """
        self._check_number_type(
            value, "time"
        )  # micropython doesn't support calling setter of parent property, do it manually
        self._time = value
        self._lat.time = value
        self._lon.time = value
        self._alt.time = value

    @SenmlRecord.update_time.setter
    def update_time(self, value):
        """set the time stamp.
        this is overridden so we can pass on the values to the internal objects.
        """
        self._check_number_type(
            value, "update_time"
        )  # micropython doesn't support calling setter of parent property, do it manually
        self._update_time = value
        self._lat.update_time = value
        self._lon.update_time = value
        self._alt.update_time = value

    @SenmlRecord._parent.setter
    def _parent(self, value):
        """set the time stamp.
        this is overridden so we can pass on the values to the internal objects.
        This is needed so that the child objects can correctly take base time (optionally also base-sum, base-value) into account
        """
        self.__parent = (
            value  # micropython doesn't support calling setter of parent property, do it manually
        )
        self._lat._parent = value
        self._lon._parent = value
        self._alt._parent = value


pack = SenmlPack("device_name")
loc = Coordinates("location")
loc2 = Coordinates("location", value=[52.0259, 5.4775, 230])
pack.add(loc)
pack.add(loc2)

print(loc._parent.name)

loc.value = [51.0259, 4.4775, 10]
print(pack.to_json())

pack.base_time = time.time()  # set a base time
time.sleep(2)
loc.time = time.time()  # all child objects will receive the time value
print(pack.to_json())



# test viper with multiple subscripts in a single expression


@micropython.viper
def f1(b: ptr8):
    b[0] += b[1]


b = bytearray(b"\x01\x02")
f1(b)
print(b)


@micropython.viper
def f2(b: ptr8, i: int):
    b[0] += b[i]


b = bytearray(b"\x01\x02")
f2(b, 1)
print(b)


@micropython.viper
def f3(b: ptr8) -> int:
    return b[0] << 24 | b[1] << 16 | b[2] << 8 | b[3]


print(hex(f3(b"\x01\x02\x03\x04")))



try:
    bytes.center
except:
    print("SKIP")
    raise SystemExit

print(b"foo".center(0))
print(b"foo".center(1))
print(b"foo".center(3))
print(b"foo".center(4))
print(b"foo".center(5))
print(b"foo".center(6))
print(b"foo".center(20))



try:
    import machine

    machine.PinBase
    machine.time_pulse_us
except:
    print("SKIP")
    raise SystemExit


class ConstPin(machine.PinBase):
    def __init__(self, value):
        self.v = value

    def value(self, v=None):
        if v is None:
            return self.v
        else:
            self.v = v


class TogglePin(machine.PinBase):
    def __init__(self):
        self.v = 0

    def value(self, v=None):
        if v is None:
            self.v = 1 - self.v
            print("value:", self.v)
            return self.v


p = TogglePin()

t = machine.time_pulse_us(p, 1)
print(type(t))
t = machine.time_pulse_us(p, 0)
print(type(t))

p = ConstPin(0)
print(machine.time_pulse_us(p, 1, 10))
print(machine.time_pulse_us(p, 0, 10))



# Test BLE GAP advertising and scanning

from micropython import const
import time, machine, bluetooth

_IRQ_SCAN_RESULT = const(5)
_IRQ_SCAN_DONE = const(6)

ADV_TIME_MS = 3000


def instance0():
    multitest.globals(BDADDR=ble.config("mac"))
    multitest.next()

    adv_data = b"\x02\x01\x06\x04\xffMPY"

    print("gap_advertise(100_000, connectable=False)")
    ble.gap_advertise(100_000, adv_data, connectable=False)
    time.sleep_ms(ADV_TIME_MS)

    print("gap_advertise(20_000, connectable=True)")
    ble.gap_advertise(20_000, adv_data, connectable=True)
    time.sleep_ms(ADV_TIME_MS)

    print("gap_advertise(None)")
    ble.gap_advertise(None)

    ble.active(0)


def instance1():
    multitest.next()
    finished = False
    matched_adv_types = {}
    matched_adv_data = None

    def irq(ev, data):
        nonlocal finished, matched_adv_types, matched_adv_data
        if ev == _IRQ_SCAN_RESULT:
            addr_type, addr, adv_type, rssi, adv_data = data
            if addr_type == BDADDR[0] and addr == BDADDR[1]:
                matched_adv_types[adv_type] = True
                if matched_adv_data is None:
                    matched_adv_data = bytes(adv_data)
                else:
                    if adv_data != matched_adv_data:
                        matched_adv_data = b"MISMATCH"
        elif ev == _IRQ_SCAN_DONE:
            finished = True

    try:
        ble.config(rxbuf=2000)
    except:
        pass
    ble.irq(irq)
    ble.gap_scan(2 * ADV_TIME_MS, 30000, 30000)
    while not finished:
        machine.idle()
    ble.active(0)
    print("adv_types:", sorted(matched_adv_types))
    print("adv_data:", matched_adv_data)


ble = bluetooth.BLE()
ble.active(1)



# import essential libraries
import pyb

lcd = pyb.LCD("x")
lcd.light(1)


# do 1 iteration of Conway's Game of Life
def conway_step():
    for x in range(128):  # loop over x coordinates
        for y in range(32):  # loop over y coordinates
            # count number of neighbours
            num_neighbours = (
                lcd.get(x - 1, y - 1)
                + lcd.get(x, y - 1)
                + lcd.get(x + 1, y - 1)
                + lcd.get(x - 1, y)
                + lcd.get(x + 1, y)
                + lcd.get(x + 1, y + 1)
                + lcd.get(x, y + 1)
                + lcd.get(x - 1, y + 1)
            )

            # check if the centre cell is alive or not
            self = lcd.get(x, y)

            # apply the rules of life
            if self and not (2 <= num_neighbours <= 3):
                lcd.pixel(x, y, 0)  # not enough, or too many neighbours: cell dies
            elif not self and num_neighbours == 3:
                lcd.pixel(x, y, 1)  # exactly 3 neighbours around an empty cell: cell is born


# randomise the start
def conway_rand():
    lcd.fill(0)  # clear the LCD
    for x in range(128):  # loop over x coordinates
        for y in range(32):  # loop over y coordinates
            lcd.pixel(x, y, pyb.rng() & 1)  # set the pixel randomly


# loop for a certain number of frames, doing iterations of Conway's Game of Life
def conway_go(num_frames):
    for i in range(num_frames):
        conway_step()  # do 1 iteration
        lcd.show()  # update the LCD
        pyb.delay(50)


# testing
conway_rand()
conway_go(100)



try:
    from collections import OrderedDict
except ImportError:
    print("SKIP")
    raise SystemExit

x = OrderedDict()
y = OrderedDict()
x['a'] = 1
x['b'] = 2
y['a'] = 1
y['b'] = 2
print(x)
print(y)
print(x == y)

z = OrderedDict()
z['b'] = 2
z['a'] = 1
print(y)
print(z)
print(y == z)

del z['b']
z['b'] = 2
print(y)
print(z)
print(y == z)

del x['a']
del y['a']
print(x)
print(y)
print(x == y)

del z['b']
del y['b']
print(y)
print(z)
print(y == z)



# test array types QqLl that require big-ints

try:
    from array import array
except ImportError:
    print("SKIP")
    raise SystemExit

print(array('L', [0, 2**32-1]))
print(array('l', [-2**31, 0, 2**31-1]))

print(array('q'))
print(array('Q'))

print(array('q', [0]))
print(array('Q', [0]))

print(array('q', [-2**63, -1, 0, 1, 2, 2**63-1]))
print(array('Q', [0, 1, 2, 2**64-1]))

print(bytes(array('q', [-1])))
print(bytes(array('Q', [2**64-1])))



try:
    from pyb import CAN
except ImportError:
    print("SKIP")
    raise SystemExit

from array import array
import micropython
import pyb

# test we can correctly create by id (2 handled in can2.py test)
for bus in (-1, 0, 1, 3):
    try:
        CAN(bus, CAN.LOOPBACK)
        print("CAN", bus)
    except ValueError:
        print("ValueError", bus)
CAN(1).deinit()

can = CAN(1)
print(can)

# Test state when de-init'd
print(can.state() == can.STOPPED)

can.init(CAN.LOOPBACK, num_filter_banks=14)
print(can)
print(can.any(0))

# Test state when freshly created
print(can.state() == can.ERROR_ACTIVE)

# Test that restart can be called
can.restart()

# Test info returns a sensible value
print(can.info())

# Catch all filter
can.setfilter(0, CAN.MASK16, 0, (0, 0, 0, 0))

can.send("abcd", 123, timeout=5000)
print(can.any(0), can.info())
print(can.recv(0))

can.send("abcd", -1, timeout=5000)
print(can.recv(0))

can.send("abcd", 0x7FF + 1, timeout=5000)
print(can.recv(0))

# Test too long message
try:
    can.send("abcdefghi", 0x7FF, timeout=5000)
except ValueError:
    print("passed")
else:
    print("failed")

# Test that recv can work without allocating memory on the heap

buf = bytearray(10)
l = [0, 0, 0, 0, memoryview(buf)]
l2 = None

micropython.heap_lock()

can.send("", 42)
l2 = can.recv(0, l)
assert l is l2
print(l, len(l[4]), buf)

can.send("1234", 42)
l2 = can.recv(0, l)
assert l is l2
print(l, len(l[4]), buf)

can.send("01234567", 42)
l2 = can.recv(0, l)
assert l is l2
print(l, len(l[4]), buf)

can.send("abc", 42)
l2 = can.recv(0, l)
assert l is l2
print(l, len(l[4]), buf)

micropython.heap_unlock()

# Test that recv can work with different arrays behind the memoryview
can.send("abc", 1)
print(bytes(can.recv(0, [0, 0, 0, 0, memoryview(array("B", range(8)))])[4]))
can.send("def", 1)
print(bytes(can.recv(0, [0, 0, 0, 0, memoryview(array("b", range(8)))])[4]))

# Test for non-list passed as second arg to recv
can.send("abc", 1)
try:
    can.recv(0, 1)
except TypeError:
    print("TypeError")

# Test for too-short-list passed as second arg to recv
can.send("abc", 1)
try:
    can.recv(0, [0, 0, 0])
except ValueError:
    print("ValueError")

# Test for non-memoryview passed as 4th element to recv
can.send("abc", 1)
try:
    can.recv(0, [0, 0, 0, 0, 0])
except TypeError:
    print("TypeError")

# Test for read-only-memoryview passed as 4th element to recv
can.send("abc", 1)
try:
    can.recv(0, [0, 0, 0, memoryview(bytes(8))])
except ValueError:
    print("ValueError")

# Test for bad-typecode-memoryview passed as 4th element to recv
can.send("abc", 1)
try:
    can.recv(0, [0, 0, 0, memoryview(array("i", range(8)))])
except ValueError:
    print("ValueError")

del can

# Testing extended IDs
print("==== TEST extframe=True ====")

can = CAN(1, CAN.LOOPBACK)
# Catch all filter, but only for extframe's
can.setfilter(0, CAN.MASK32, 0, (0, 0), extframe=True)

print(can)

try:
    can.send("abcde", 0x7FF + 1, timeout=5000, extframe=True)
except ValueError:
    print("failed")
else:
    r = can.recv(0)
    if r[0] == 0x7FF + 1 and r[4] == b"abcde":
        print("passed")
    else:
        print("failed, wrong data received")

# Test filters
for n in [0, 8, 16, 24]:
    filter_id = 0b00001000 << n
    filter_mask = 0b00011100 << n
    id_ok = 0b00001010 << n
    id_fail = 0b00011010 << n

    can.clearfilter(0, extframe=True)
    can.setfilter(0, pyb.CAN.MASK32, 0, (filter_id, filter_mask), extframe=True)

    can.send("ok", id_ok, timeout=3, extframe=True)
    if can.any(0):
        msg = can.recv(0)
        print((hex(filter_id), hex(filter_mask), hex(msg[0]), msg[1], msg[4]))

    can.send("fail", id_fail, timeout=3, extframe=True)
    if can.any(0):
        msg = can.recv(0)
        print((hex(filter_id), hex(filter_mask), hex(msg[0]), msg[1], msg[4]))

del can

# Test RxCallbacks
print("==== TEST rx callbacks ====")

can = CAN(1, CAN.LOOPBACK)
can.setfilter(0, CAN.LIST16, 0, (1, 2, 3, 4))
can.setfilter(1, CAN.LIST16, 1, (5, 6, 7, 8))


def cb0(bus, reason):
    print("cb0")
    if reason == 0:
        print("pending")
    if reason == 1:
        print("full")
    if reason == 2:
        print("overflow")


def cb1(bus, reason):
    print("cb1")
    if reason == 0:
        print("pending")
    if reason == 1:
        print("full")
    if reason == 2:
        print("overflow")


def cb0a(bus, reason):
    print("cb0a")
    if reason == 0:
        print("pending")
    if reason == 1:
        print("full")
    if reason == 2:
        print("overflow")


def cb1a(bus, reason):
    print("cb1a")
    if reason == 0:
        print("pending")
    if reason == 1:
        print("full")
    if reason == 2:
        print("overflow")


can.rxcallback(0, cb0)
can.rxcallback(1, cb1)

can.send("11111111", 1, timeout=5000)
can.send("22222222", 2, timeout=5000)
can.send("33333333", 3, timeout=5000)
can.rxcallback(0, cb0a)
can.send("44444444", 4, timeout=5000)

can.send("55555555", 5, timeout=5000)
can.send("66666666", 6, timeout=5000)
can.send("77777777", 7, timeout=5000)
can.rxcallback(1, cb1a)
can.send("88888888", 8, timeout=5000)

print(can.recv(0))
print(can.recv(0))
print(can.recv(0))
print(can.recv(1))
print(can.recv(1))
print(can.recv(1))

can.send("11111111", 1, timeout=5000)
can.send("55555555", 5, timeout=5000)

print(can.recv(0))
print(can.recv(1))

del can

# Testing asynchronous send
print("==== TEST async send ====")

can = CAN(1, CAN.LOOPBACK)
can.setfilter(0, CAN.MASK16, 0, (0, 0, 0, 0))

while can.any(0):
    can.recv(0)

can.send("abcde", 1, timeout=0)
print(can.any(0))
while not can.any(0):
    pass

print(can.recv(0))

try:
    can.send("abcde", 2, timeout=0)
    can.send("abcde", 3, timeout=0)
    can.send("abcde", 4, timeout=0)
    can.send("abcde", 5, timeout=0)
except OSError as e:
    if str(e) == "16":
        print("passed")
    else:
        print("failed")

pyb.delay(500)
while can.any(0):
    print(can.recv(0))

# Testing rtr messages
print("==== TEST rtr messages ====")

bus1 = CAN(1, CAN.LOOPBACK)
while bus1.any(0):
    bus1.recv(0)
bus1.setfilter(0, CAN.LIST16, 0, (1, 2, 3, 4))
bus1.setfilter(1, CAN.LIST16, 0, (5, 6, 7, 8), rtr=(True, True, True, True))
bus1.setfilter(2, CAN.MASK16, 0, (64, 64, 32, 32), rtr=(False, True))

bus1.send("", 1, rtr=True)
print(bus1.any(0))
bus1.send("", 5, rtr=True)
print(bus1.recv(0))
bus1.send("", 6, rtr=True)
print(bus1.recv(0))
bus1.send("", 7, rtr=True)
print(bus1.recv(0))
bus1.send("", 16, rtr=True)
print(bus1.any(0))
bus1.send("", 32, rtr=True)
print(bus1.recv(0))

# test HAL error, timeout
print("==== TEST errors ====")

can = pyb.CAN(1, pyb.CAN.NORMAL)
try:
    can.send("1", 1, timeout=50)
except OSError as e:
    print(repr(e))



import sys
import struct
import hashlib

# This region at the start of flash contains a small header and then segments
# containing .text, .data and .rodata, and so must be large enough to hold all
# of this.  This data is loaded to the appropriate places in RAM by the ROM
# bootloader at boot.  After this in flash comes .irom0.text, which must begin
# on a flash erase-page boundary.
SEGS_MAX_SIZE = 0x9000

assert len(sys.argv) == 4

md5 = hashlib.md5()

with open(sys.argv[3], "wb") as fout:
    with open(sys.argv[1], "rb") as f:
        data_flash = f.read()
        fout.write(data_flash)
        # First 4 bytes include flash size, etc. which may be changed
        # by esptool.py, etc.
        md5.update(data_flash[4:])
        print("flash    ", len(data_flash))

        # Print info about segments in this first part of flash
        num_segs = struct.unpack_from("<BBBBI", data_flash, 0)[1]
        offset = 8
        for seg_num in range(num_segs):
            seg_name = [".text", ".data", ".rodata"][seg_num]
            seg_offset, seg_size = struct.unpack_from("<II", data_flash, offset)
            print(" {:7}  {} at 0x{:x}".format(seg_name, seg_size, seg_offset))
            offset += 8 + seg_size

    with open(sys.argv[2], "rb") as f:
        data_rom = f.read()

    pad = b"\xff" * (SEGS_MAX_SIZE - len(data_flash))
    assert len(pad) >= 4
    fout.write(pad[:-4])
    md5.update(pad[:-4])
    len_data = struct.pack("I", SEGS_MAX_SIZE + len(data_rom))
    fout.write(len_data)
    md5.update(len_data)
    print("padding  ", len(pad))

    fout.write(data_rom)
    md5.update(data_rom)
    print("irom0text", len(data_rom))

    fout.write(md5.digest())

    print("total    ", SEGS_MAX_SIZE + len(data_rom))
    print("md5      ", md5.hexdigest())



#!/usr/bin/env python3
"""
Command-line interface for the NEAR Python contract compiler.
"""

import shutil
import sys
from pathlib import Path
from typing import Optional

import rich_click as click

from .builder import compile_contract
from .utils import console, is_running_in_container, setup_venv


def find_contract_file() -> Optional[Path]:
    """
    Automatically detect a contract file in the current directory.
    Looks for __init__.py first, then main.py.

    Returns:
        Path to the contract file if found, None otherwise
    """
    current_dir = Path.cwd()

    # Check for __init__.py
    init_path = current_dir / "__init__.py"
    if init_path.exists() and init_path.is_file():
        return init_path

    # Check for main.py
    main_path = current_dir / "main.py"
    if main_path.exists() and main_path.is_file():
        return main_path

    # No contract file found
    return None


@click.command()
@click.argument(
    "contract", type=click.Path(exists=True, dir_okay=False), required=False
)
@click.option("--output", "-o", help="Output WASM file path")
@click.option("--venv", help="Path to virtual environment", default=".venv")
@click.option("--rebuild", is_flag=True, help="Force a clean rebuild")
@click.option(
    "--reproducible", is_flag=True, help="Build reproducibly in Docker container"
)
@click.option(
    "--init-reproducible-config",
    is_flag=True,
    help="Initialize reproducible build configuration in pyproject.toml",
)
@click.option(
    "--single-file",
    is_flag=True,
    help="Skip local module discovery, compile only the specified file",
)
@click.option(
    "--create-venv",
    is_flag=True,
    help="Force setup of virtual environment before building",
)
def main(
    contract: Optional[str],
    output: Optional[str],
    venv: str,
    rebuild: bool,
    reproducible: bool,
    init_reproducible_config: bool,
    single_file: bool,
    create_venv: bool,
):
    """Compile a Python contract to WebAssembly for NEAR blockchain.

    If CONTRACT is not specified, looks for __init__.py or main.py in the current directory.
    """
    # Handle initialization of reproducible build configuration
    if init_reproducible_config:
        from .reproducible import init_reproducible_build_config

        current_dir = Path.cwd()
        if init_reproducible_build_config(current_dir):
            console.print("[green]Reproducible build configuration initialized")
        else:
            console.print("[red]Failed to initialize reproducible build configuration")
        sys.exit(0)

    # Try to auto-detect contract file if not provided
    contract_path = None
    if contract:
        contract_path = Path(contract).resolve()
    else:
        detected_contract = find_contract_file()
        if detected_contract:
            contract_path = detected_contract.resolve()
            console.print(
                f"[cyan]Auto-detected contract file:[/] [yellow]{contract_path}[/]"
            )
        else:
            console.print(
                "[red]Error: No contract file specified and could not auto-detect __init__.py or main.py"
            )
            console.print(
                "[cyan]Please specify a contract file or create an __init__.py or main.py file"
            )
            sys.exit(1)

    venv_path = Path(venv).resolve()
    contract_dir = contract_path.parent

    # Determine output path if not specified
    if not output:
        output = f"{contract_path.stem}.wasm"
    output_path = Path(output).resolve()

    # If reproducible flag is set, build in Docker
    if reproducible:
        from .reproducible import run_reproducible_build

        # Prepare build args
        build_args = []
        if rebuild:
            build_args.append("--rebuild")
        if single_file:
            build_args.append("--single-file")

        # Run reproducible build in Docker
        if not run_reproducible_build(contract_path, output_path, build_args):
            console.print("[red]Failed to run reproducible build")
            sys.exit(1)

        sys.exit(0)

    # Check for container environment and set up venv if needed
    in_container = is_running_in_container()
    if in_container:
        console.print(
            "[cyan]Detected running in container, setting up environment automatically[/]"
        )
        if not setup_venv(venv_path, contract_dir):
            console.print("[red]Failed to set up virtual environment in container")
            sys.exit(1)
    elif create_venv:
        # User explicitly requested venv setup
        if not setup_venv(venv_path, contract_dir):
            console.print("[red]Failed to set up virtual environment")
            sys.exit(1)
    elif not venv_path.exists():
        console.print(f"[red]Error: Virtual environment not found at {venv_path}")
        console.print("[cyan]Create one with: uv init")
        console.print("[cyan]Or run with --create-venv to create it automatically")
        sys.exit(1)

    # Check that emcc is available
    if not shutil.which("emcc"):
        console.print("[red]Error: Emscripten compiler (emcc) not found in PATH")
        console.print(
            "[cyan]Please install Emscripten: https://emscripten.org/docs/getting_started/"
        )
        sys.exit(1)

    # Determine assets directory
    assets_dir = Path(__file__).parent
    if not (assets_dir / "micropython").exists():
        console.print(
            f"[red]Error: MicroPython assets not found at {assets_dir / 'micropython'}"
        )
        sys.exit(1)

    # Compile the contract
    if not compile_contract(
        contract_path, output_path, venv_path, assets_dir, rebuild, single_file
    ):
        sys.exit(1)


if __name__ == "__main__":
    main()



try:
    import uctypes
except ImportError:
    print("SKIP")
    raise SystemExit

desc = {
    # arr is array at offset 0, of UINT8 elements, array size is 2
    "arr": (uctypes.ARRAY | 0, uctypes.UINT8 | 2),
    # arr2 is array at offset 0, size 2, of structures defined recursively
    "arr2": (uctypes.ARRAY | 0, 2, {"b": uctypes.UINT8 | 0}),
    "arr3": (uctypes.ARRAY | 2, uctypes.UINT16 | 2),
    "arr4": (uctypes.ARRAY | 0, 2, {"b": uctypes.UINT8 | 0, "w": uctypes.UINT16 | 1}),
    "sub": (
        0,
        {
            "b1": uctypes.BFUINT8 | 0 | 4 << uctypes.BF_POS | 4 << uctypes.BF_LEN,
            "b2": uctypes.BFUINT8 | 0 | 0 << uctypes.BF_POS | 4 << uctypes.BF_LEN,
        },
    ),
}

data = bytearray(b"01234567")

S = uctypes.struct(uctypes.addressof(data), desc, uctypes.LITTLE_ENDIAN)

print(uctypes.sizeof(S.arr))
assert uctypes.sizeof(S.arr) == 2

print(uctypes.sizeof(S.arr2))
assert uctypes.sizeof(S.arr2) == 2

print(uctypes.sizeof(S.arr3))

try:
    print(uctypes.sizeof(S.arr3[0]))
except TypeError:
    print("TypeError")

print(uctypes.sizeof(S.arr4))
assert uctypes.sizeof(S.arr4) == 6

print(uctypes.sizeof(S.sub))
assert uctypes.sizeof(S.sub) == 1

# invalid descriptors
try:
    print(uctypes.sizeof([]))
except TypeError:
    print("TypeError")

try:
    print(uctypes.sizeof(()))
except TypeError:
    print("TypeError")

try:
    print(uctypes.sizeof(("garbage",)))
except TypeError:
    print("TypeError")

try:
    # PTR * 3 is intended to be an invalid agg_type (STRUCT, PTR, ARRAY are valid ones).
    print(uctypes.sizeof((uctypes.PTR * 3,)))
except TypeError:
    print("TypeError")

try:
    print(uctypes.sizeof((0, {}, "garbage")))
except TypeError:
    print("TypeError")

try:
    print(uctypes.sizeof((uctypes.PTR | 0,)))
except TypeError:
    print("TypeError")

try:
    print(uctypes.sizeof((uctypes.ARRAY | 0,)))
except TypeError:
    print("TypeError")

try:
    print(uctypes.sizeof((uctypes.ARRAY | 0, 1, {}, "garbage")))
except TypeError:
    print("TypeError")

# empty descriptor
print(uctypes.sizeof({}))



import sys
import os


class LS:
    def __repr__(self):
        self.__call__()
        return ""

    def __call__(self, path="."):
        l = list(os.ilistdir(path))
        l.sort()
        for f in l:
            if f[1] == 0x4000:  # stat.S_IFDIR
                print("    <dir> %s" % f[0])
        for f in l:
            if f[1] != 0x4000:
                if len(f) > 3:
                    print("% 9d %s" % (f[3], f[0]))
                else:
                    print("          %s" % f[0])
        try:
            st = os.statvfs(path)
            print("\n{:,d}k free".format(st[1] * st[3] // 1024))
        except:
            pass


class PWD:
    def __repr__(self):
        return os.getcwd()

    def __call__(self):
        return self.__repr__()


class CLEAR:
    def __repr__(self):
        return "\x1b[2J\x1b[H"

    def __call__(self):
        return self.__repr__()


def head(f, n=10):
    with open(f) as f:
        for i in range(n):
            l = f.readline()
            if not l:
                break
            sys.stdout.write(l)


def cat(f):
    head(f, 1 << 30)


def cp(s, t):
    try:
        if os.stat(t)[0] & 0x4000:  # is directory
            t = t.rstrip("/") + "/" + s
    except OSError:
        pass
    buf = bytearray(512)
    buf_mv = memoryview(buf)
    with open(s, "rb") as s, open(t, "wb") as t:
        while True:
            n = s.readinto(buf)
            if n <= 0:
                break
            t.write(buf_mv[:n])


def newfile(path):
    print("Type file contents line by line, finish with EOF (Ctrl+D).")
    with open(path, "w") as f:
        while 1:
            try:
                l = input()
            except EOFError:
                break
            f.write(l)
            f.write("\n")


def rm(d, recursive=False):  # Remove file or tree
    try:
        if (os.stat(d)[0] & 0x4000) and recursive:  # Dir
            for f in os.ilistdir(d):
                if f[0] != "." and f[0] != "..":
                    rm("/".join((d, f[0])))  # File or Dir
            os.rmdir(d)
        else:  # File
            os.remove(d)
    except:
        print("rm of '%s' failed" % d)


class Man:
    def __repr__(self):
        return """
upysh is intended to be imported using:
from upysh import *

To see this help text again, type "man".

upysh commands:
clear, ls, ls(...), head(...), cat(...), newfile(...)
cp('src', 'dest'), mv('old', 'new'), rm(...)
pwd, cd(...), mkdir(...), rmdir(...)
"""


man = Man()
pwd = PWD()
ls = LS()
clear = CLEAR()

cd = os.chdir
mkdir = os.mkdir
mv = os.rename
rmdir = os.rmdir

print(man)



# MicroPython USB host library, compatible with PyUSB.
# MIT license; Copyright (c) 2021-2024 Damien P. George

import sys
import ffi
import uctypes

if sys.maxsize >> 32:
    UINTPTR_SIZE = 8
    UINTPTR = uctypes.UINT64
else:
    UINTPTR_SIZE = 4
    UINTPTR = uctypes.UINT32


def _align_word(x):
    return (x + UINTPTR_SIZE - 1) & ~(UINTPTR_SIZE - 1)


ptr_descriptor = (0 | uctypes.ARRAY, 1 | UINTPTR)

libusb_device_descriptor = {
    "bLength": 0 | uctypes.UINT8,
    "bDescriptorType": 1 | uctypes.UINT8,
    "bcdUSB": 2 | uctypes.UINT16,
    "bDeviceClass": 4 | uctypes.UINT8,
    "bDeviceSubClass": 5 | uctypes.UINT8,
    "bDeviceProtocol": 6 | uctypes.UINT8,
    "bMaxPacketSize0": 7 | uctypes.UINT8,
    "idVendor": 8 | uctypes.UINT16,
    "idProduct": 10 | uctypes.UINT16,
    "bcdDevice": 12 | uctypes.UINT16,
    "iManufacturer": 14 | uctypes.UINT8,
    "iProduct": 15 | uctypes.UINT8,
    "iSerialNumber": 16 | uctypes.UINT8,
    "bNumConfigurations": 17 | uctypes.UINT8,
}

libusb_config_descriptor = {
    "bLength": 0 | uctypes.UINT8,
    "bDescriptorType": 1 | uctypes.UINT8,
    "wTotalLength": 2 | uctypes.UINT16,
    "bNumInterfaces": 4 | uctypes.UINT8,
    "bConfigurationValue": 5 | uctypes.UINT8,
    "iConfiguration": 6 | uctypes.UINT8,
    "bmAttributes": 7 | uctypes.UINT8,
    "MaxPower": 8 | uctypes.UINT8,
    "interface": _align_word(9) | UINTPTR,  # array of libusb_interface
    "extra": (_align_word(9) + UINTPTR_SIZE) | UINTPTR,
    "extra_length": (_align_word(9) + 2 * UINTPTR_SIZE) | uctypes.INT,
}

libusb_interface = {
    "altsetting": 0 | UINTPTR,  # array of libusb_interface_descriptor
    "num_altsetting": UINTPTR_SIZE | uctypes.INT,
}

libusb_interface_descriptor = {
    "bLength": 0 | uctypes.UINT8,
    "bDescriptorType": 1 | uctypes.UINT8,
    "bInterfaceNumber": 2 | uctypes.UINT8,
    "bAlternateSetting": 3 | uctypes.UINT8,
    "bNumEndpoints": 4 | uctypes.UINT8,
    "bInterfaceClass": 5 | uctypes.UINT8,
    "bInterfaceSubClass": 6 | uctypes.UINT8,
    "bInterfaceProtocol": 7 | uctypes.UINT8,
    "iInterface": 8 | uctypes.UINT8,
    "endpoint": _align_word(9) | UINTPTR,
    "extra": (_align_word(9) + UINTPTR_SIZE) | UINTPTR,
    "extra_length": (_align_word(9) + 2 * UINTPTR_SIZE) | uctypes.INT,
}

libusb = ffi.open("libusb-1.0.so")
libusb_init = libusb.func("i", "libusb_init", "p")
libusb_exit = libusb.func("v", "libusb_exit", "p")
libusb_get_device_list = libusb.func("i", "libusb_get_device_list", "pp")  # return is ssize_t
libusb_free_device_list = libusb.func("v", "libusb_free_device_list", "pi")
libusb_get_device_descriptor = libusb.func("i", "libusb_get_device_descriptor", "pp")
libusb_get_config_descriptor = libusb.func("i", "libusb_get_config_descriptor", "pBp")
libusb_free_config_descriptor = libusb.func("v", "libusb_free_config_descriptor", "p")
libusb_open = libusb.func("i", "libusb_open", "pp")
libusb_set_configuration = libusb.func("i", "libusb_set_configuration", "pi")
libusb_claim_interface = libusb.func("i", "libusb_claim_interface", "pi")
libusb_control_transfer = libusb.func("i", "libusb_control_transfer", "pBBHHpHI")


def _new(sdesc):
    buf = bytearray(uctypes.sizeof(sdesc))
    s = uctypes.struct(uctypes.addressof(buf), sdesc)
    return s


class Interface:
    def __init__(self, descr):
        # Public attributes.
        self.bInterfaceClass = descr.bInterfaceClass
        self.bInterfaceSubClass = descr.bInterfaceSubClass
        self.iInterface = descr.iInterface
        self.extra_descriptors = uctypes.bytes_at(descr.extra, descr.extra_length)


class Configuration:
    def __init__(self, dev, cfg_idx):
        cfgs = _new(ptr_descriptor)
        if libusb_get_config_descriptor(dev._dev, cfg_idx, cfgs) != 0:
            libusb_exit(0)
            raise Exception
        descr = uctypes.struct(cfgs[0], libusb_config_descriptor)

        # Extract all needed info because descr is going to be free'd at the end.
        self._itfs = []
        itf_array = uctypes.struct(
            descr.interface, (0 | uctypes.ARRAY, descr.bNumInterfaces, libusb_interface)
        )
        for i in range(descr.bNumInterfaces):
            itf = itf_array[i]
            alt_array = uctypes.struct(
                itf.altsetting,
                (0 | uctypes.ARRAY, itf.num_altsetting, libusb_interface_descriptor),
            )
            for j in range(itf.num_altsetting):
                alt = alt_array[j]
                self._itfs.append(Interface(alt))

        # Public attributes.
        self.bNumInterfaces = descr.bNumInterfaces
        self.bConfigurationValue = descr.bConfigurationValue
        self.bmAttributes = descr.bmAttributes
        self.bMaxPower = descr.MaxPower
        self.extra_descriptors = uctypes.bytes_at(descr.extra, descr.extra_length)

        # Free descr memory in the driver.
        libusb_free_config_descriptor(cfgs[0])

    def __iter__(self):
        return iter(self._itfs)


class Device:
    _TIMEOUT_DEFAULT = 1000

    def __init__(self, dev, descr):
        self._dev = dev
        self._num_cfg = descr.bNumConfigurations
        self._handle = None
        self._claim_itf = set()

        # Public attributes.
        self.idVendor = descr.idVendor
        self.idProduct = descr.idProduct

    def __iter__(self):
        for i in range(self._num_cfg):
            yield Configuration(self, i)

    def __getitem__(self, i):
        return Configuration(self, i)

    def _open(self):
        if self._handle is None:
            # Open the USB device.
            handle = _new(ptr_descriptor)
            if libusb_open(self._dev, handle) != 0:
                libusb_exit(0)
                raise Exception
            self._handle = handle[0]

    def _claim_interface(self, i):
        if libusb_claim_interface(self._handle, i) != 0:
            libusb_exit(0)
            raise Exception

    def set_configuration(self):
        # Select default configuration.
        self._open()
        cfg = Configuration(self, 0).bConfigurationValue
        ret = libusb_set_configuration(self._handle, cfg)
        if ret != 0:
            libusb_exit(0)
            raise Exception

    def ctrl_transfer(
        self, bmRequestType, bRequest, wValue=0, wIndex=0, data_or_wLength=None, timeout=None
    ):
        if data_or_wLength is None:
            l = 0
            data = bytes()
        elif isinstance(data_or_wLength, int):
            l = data_or_wLength
            data = bytearray(l)
        else:
            l = len(data_or_wLength)
            data = data_or_wLength
        self._open()
        if wIndex & 0xFF not in self._claim_itf:
            self._claim_interface(wIndex & 0xFF)
            self._claim_itf.add(wIndex & 0xFF)
        if timeout is None:
            timeout = self._TIMEOUT_DEFAULT
        ret = libusb_control_transfer(
            self._handle, bmRequestType, bRequest, wValue, wIndex, data, l, timeout * 1000
        )
        if ret < 0:
            libusb_exit(0)
            raise Exception
        if isinstance(data_or_wLength, int):
            return data
        else:
            return ret


def find(*, find_all=False, custom_match=None, idVendor=None, idProduct=None):
    if libusb_init(0) < 0:
        raise Exception

    devs = _new(ptr_descriptor)
    count = libusb_get_device_list(0, devs)
    if count < 0:
        libusb_exit(0)
        raise Exception

    dev_array = uctypes.struct(devs[0], (0 | uctypes.ARRAY, count | UINTPTR))
    descr = _new(libusb_device_descriptor)
    devices = None
    for i in range(count):
        libusb_get_device_descriptor(dev_array[i], descr)
        if idVendor and descr.idVendor != idVendor:
            continue
        if idProduct and descr.idProduct != idProduct:
            continue
        device = Device(dev_array[i], descr)
        if custom_match and not custom_match(device):
            continue
        if not find_all:
            return device
        if not devices:
            devices = []
        devices.append(device)
    return devices



# See https://github.com/python/cpython/blob/3.9/Lib/test/datetimetester.py
#
# This script can be run in 3 different modes:
# 1. `python3 test_datetime.py --stdlib`: checks that the tests comply to
#    CPython's standard datetime library.
# 2. `python3 test_datetime.py`: runs the tests against datetime.py, using
#    CPython's standard unittest (which accepts filter options, such as
#    `-v TestTimeDelta -k tuple`, and provides more verbose output in case
#    of failure).
# 3. `micropython test_datetime.py`: runs the tests against datetime.py
#    using MicroPython's unittest library (which must be available).
#
# This script also accepts option `--reorder` which rewrites this file
# in-place by numbering tests in sequence.

import sys

STDLIB = False

if __name__ == "__main__":
    while len(sys.argv) > 1:
        if sys.argv[1] == "--reorder":
            import fileinput, re

            with fileinput.input(files=sys.argv[0], inplace=True) as f:
                cases = {}
                n = 0
                for line in f:
                    match = re.match("(\s+def\s+test_)(\w+?)(?:\d+)(\(.+\):)", line)
                    if match:
                        prefix, name, suffix = match.groups()
                        if name != last_name:
                            if name in cases[case]:
                                sys.exit(
                                    f"duplicated test in {case} at line {fileinput.filelineno()}: {name}"
                                )
                            cases[case].append(name)
                            last_name = name
                            i = 0
                        print(f"{prefix}{name}{i:02d}{suffix}")
                        i += 1
                        n += 1
                        continue

                    match = re.match("class\s+(Test[\w\d]+)\(", line)
                    if match:
                        case = match[1]
                        if case in cases:
                            sys.exit(
                                f"duplicated test case at line {fileinput.filelineno()}: {case}"
                            )
                        cases[case] = []
                        last_name = ""

                    print(line, end="")
            print(f"Reordered {n} tests in {len(cases)} cases")
        elif sys.argv[1] == "--stdlib":
            sys.path.pop(0)
            STDLIB = True
        else:
            break
        sys.argv.pop(1)

import os
import time as mod_time
import datetime as mod_datetime
from datetime import MAXYEAR, MINYEAR, datetime, date, time, timedelta, timezone, tzinfo
import unittest


# See localtz.patch
try:
    datetime.fromtimestamp(0)
    LOCALTZ = True
except NotImplementedError:
    LOCALTZ = False


if hasattr(datetime, "EPOCH"):
    EPOCH = datetime.EPOCH
else:
    EPOCH = datetime(*mod_time.gmtime(0)[:6], tzinfo=timezone.utc)


def eval_mod(s):
    return eval(s.replace("datetime.", "mod_datetime."))


### timedelta ################################################################

a = timedelta(hours=7)
b = timedelta(minutes=6)
c = timedelta(seconds=10)
us = timedelta(microseconds=1)
td0 = timedelta(0)
td1 = timedelta(2, 3, 4)
td2 = timedelta(2, 3, 4)
td3 = timedelta(2, 3, 5)
td4 = timedelta(
    days=100,
    weeks=-7,
    hours=-24 * (100 - 49),
    minutes=-3,
    seconds=12,
    microseconds=(3 * 60 - 12) * 1000000,
)  # == timedelta(0)

td1h = timedelta(hours=1)
td1hr = "datetime.timedelta(microseconds={})".format(1 * 3600 * 10**6)
td10h2m = timedelta(hours=10, minutes=2)
td10h2mr = "datetime.timedelta(microseconds={})".format((10 * 3600 + 2 * 60) * 10**6)
tdn10h2m40s = timedelta(hours=-10, minutes=2, seconds=40)
tdn10h2m40sr = "datetime.timedelta(microseconds={})".format((-10 * 3600 + 2 * 60 + 40) * 10**6)
td1h2m40s100us = timedelta(hours=1, minutes=2, seconds=40, microseconds=100)
td1h2m40s100usr = "datetime.timedelta(microseconds={})".format(
    (1 * 3600 + 2 * 60 + 40) * 10**6 + 100
)


class Test0TimeDelta(unittest.TestCase):
    def test___init__00(self):
        self.assertEqual(timedelta(), timedelta(weeks=0, days=0, hours=0, minutes=0, seconds=0))

    def test___init__01(self):
        self.assertEqual(timedelta(weeks=1), timedelta(days=7))

    def test___init__02(self):
        self.assertEqual(timedelta(days=1), timedelta(hours=24))

    def test___init__03(self):
        self.assertEqual(timedelta(hours=1), timedelta(minutes=60))

    def test___init__04(self):
        self.assertEqual(timedelta(minutes=1), timedelta(seconds=60))

    def test___init__05(self):
        self.assertEqual(timedelta(seconds=1), timedelta(milliseconds=1000))

    def test___init__06(self):
        self.assertEqual(timedelta(milliseconds=1), timedelta(microseconds=1000))

    def test___init__07(self):
        self.assertEqual(timedelta(weeks=1.0 / 7), timedelta(days=1))

    def test___init__08(self):
        self.assertEqual(timedelta(days=1.0 / 24), timedelta(hours=1))

    def test___init__09(self):
        self.assertEqual(timedelta(hours=1.0 / 60), timedelta(minutes=1))

    def test___init__10(self):
        self.assertEqual(timedelta(minutes=1.0 / 60), timedelta(seconds=1))

    def test___init__11(self):
        self.assertEqual(timedelta(seconds=0.001), timedelta(milliseconds=1))

    def test___init__12(self):
        self.assertEqual(timedelta(milliseconds=0.001), timedelta(microseconds=1))

    def test___init__13(self):
        self.assertEqual(td1h, eval_mod(td1hr))

    def test___init__14(self):
        self.assertEqual(td10h2m, eval_mod(td10h2mr))

    def test___init__15(self):
        self.assertEqual(tdn10h2m40s, eval_mod(tdn10h2m40sr))

    def test___init__16(self):
        self.assertEqual(td1h2m40s100us, eval_mod(td1h2m40s100usr))

    @unittest.skipIf(STDLIB, "standard datetime differs")
    def test___repr__00(self):
        self.assertEqual(repr(td1h), td1hr)

    @unittest.skipIf(STDLIB, "standard datetime differs")
    def test___repr__01(self):
        self.assertEqual(repr(td10h2m), td10h2mr)

    @unittest.skipIf(STDLIB, "standard datetime differs")
    def test___repr__02(self):
        self.assertEqual(repr(tdn10h2m40s), tdn10h2m40sr)

    @unittest.skipIf(STDLIB, "standard datetime differs")
    def test___repr__03(self):
        self.assertEqual(repr(td1h2m40s100us), td1h2m40s100usr)

    def test___repr__04(self):
        self.assertEqual(td1, eval_mod(repr(td1)))

    def test_total_seconds00(self):
        d = timedelta(days=365)
        self.assertEqual(d.total_seconds(), 31536000.0)

    def test_days00(self):
        self.assertEqual(td1.days, 2)

    def test_seconds00(self):
        self.assertEqual(td1.seconds, 3)

    def test_microseconds00(self):
        self.assertEqual(td1.microseconds, 4)

    def test___add__00(self):
        self.assertEqual(a + b + c, timedelta(hours=7, minutes=6, seconds=10))

    def test___add__01(self):
        dt = a + datetime(2010, 1, 1, 12, 30)
        self.assertEqual(dt, datetime(2010, 1, 1, 12 + 7, 30))

    def test___sub__00(self):
        self.assertEqual(a - b, timedelta(hours=6, minutes=60 - 6))

    def test___neg__00(self):
        self.assertEqual(-a, timedelta(hours=-7))

    def test___neg__01(self):
        self.assertEqual(-b, timedelta(hours=-1, minutes=54))

    def test___neg__02(self):
        self.assertEqual(-c, timedelta(hours=-1, minutes=59, seconds=50))

    def test___pos__00(self):
        self.assertEqual(+a, timedelta(hours=7))

    def test___abs__00(self):
        self.assertEqual(abs(a), a)

    def test___abs__01(self):
        self.assertEqual(abs(-a), a)

    def test___mul__00(self):
        self.assertEqual(a * 10, timedelta(hours=70))

    def test___mul__01(self):
        self.assertEqual(a * 10, 10 * a)

    def test___mul__02(self):
        self.assertEqual(b * 10, timedelta(minutes=60))

    def test___mul__03(self):
        self.assertEqual(10 * b, timedelta(minutes=60))

    def test___mul__04(self):
        self.assertEqual(c * 10, timedelta(seconds=100))

    def test___mul__05(self):
        self.assertEqual(10 * c, timedelta(seconds=100))

    def test___mul__06(self):
        self.assertEqual(a * -1, -a)

    def test___mul__07(self):
        self.assertEqual(b * -2, -b - b)

    def test___mul__08(self):
        self.assertEqual(c * -2, -c + -c)

    def test___mul__09(self):
        self.assertEqual(b * (60 * 24), (b * 60) * 24)

    def test___mul__10(self):
        self.assertEqual(b * (60 * 24), (60 * b) * 24)

    def test___mul__11(self):
        self.assertEqual(c * 6, timedelta(minutes=1))

    def test___mul__12(self):
        self.assertEqual(6 * c, timedelta(minutes=1))

    def test___truediv__00(self):
        self.assertEqual(a / 0.5, timedelta(hours=14))

    def test___truediv__01(self):
        self.assertEqual(b / 0.5, timedelta(minutes=12))

    def test___truediv__02(self):
        self.assertEqual(a / 7, timedelta(hours=1))

    def test___truediv__03(self):
        self.assertEqual(b / 6, timedelta(minutes=1))

    def test___truediv__04(self):
        self.assertEqual(c / 10, timedelta(seconds=1))

    def test___truediv__05(self):
        self.assertEqual(a / 10, timedelta(minutes=7 * 6))

    def test___truediv__06(self):
        self.assertEqual(a / 3600, timedelta(seconds=7))

    def test___truediv__07(self):
        self.assertEqual(a / a, 1.0)

    def test___truediv__08(self):
        t = timedelta(hours=1, minutes=24, seconds=19)
        second = timedelta(seconds=1)
        self.assertEqual(t / second, 5059.0)

    def test___truediv__09(self):
        t = timedelta(minutes=2, seconds=30)
        minute = timedelta(minutes=1)
        self.assertEqual(t / minute, 2.5)

    def test___floordiv__00(self):
        self.assertEqual(a // 7, timedelta(hours=1))

    def test___floordiv__01(self):
        self.assertEqual(b // 6, timedelta(minutes=1))

    def test___floordiv__02(self):
        self.assertEqual(c // 10, timedelta(seconds=1))

    def test___floordiv__03(self):
        self.assertEqual(a // 10, timedelta(minutes=7 * 6))

    def test___floordiv__04(self):
        self.assertEqual(a // 3600, timedelta(seconds=7))

    def test___floordiv__05(self):
        t = timedelta(hours=1, minutes=24, seconds=19)
        second = timedelta(seconds=1)
        self.assertEqual(t // second, 5059)

    def test___floordiv__06(self):
        t = timedelta(minutes=2, seconds=30)
        minute = timedelta(minutes=1)
        self.assertEqual(t // minute, 2)

    def test___mod__00(self):
        t = timedelta(minutes=2, seconds=30)
        r = t % timedelta(minutes=1)
        self.assertEqual(r, timedelta(seconds=30))

    def test___mod__01(self):
        t = timedelta(minutes=-2, seconds=30)
        r = t % timedelta(minutes=1)
        self.assertEqual(r, timedelta(seconds=30))

    def test___divmod__00(self):
        t = timedelta(minutes=2, seconds=30)
        q, r = divmod(t, timedelta(minutes=1))
        self.assertEqual(q, 2)
        self.assertEqual(r, timedelta(seconds=30))

    def test___divmod__01(self):
        t = timedelta(minutes=-2, seconds=30)
        q, r = divmod(t, timedelta(minutes=1))
        self.assertEqual(q, -2)
        self.assertEqual(r, timedelta(seconds=30))

    def test___eq__00(self):
        self.assertEqual(td1, td2)

    def test___eq__01(self):
        self.assertTrue(not td1 != td2)

    def test___eq__02(self):
        self.assertEqual(timedelta(hours=6, minutes=60), a)

    def test___eq__03(self):
        self.assertEqual(timedelta(seconds=60 * 6), b)

    def test___eq__04(self):
        self.assertTrue(not td1 == td3)

    def test___eq__05(self):
        self.assertTrue(td1 != td3)

    def test___eq__06(self):
        self.assertTrue(td3 != td1)

    def test___eq__07(self):
        self.assertTrue(not td3 == td1)

    def test___le__00(self):
        self.assertTrue(td1 <= td2)

    def test___le__01(self):
        self.assertTrue(td1 <= td3)

    def test___le__02(self):
        self.assertTrue(not td3 <= td1)

    def test___lt__00(self):
        self.assertTrue(not td1 < td2)

    def test___lt__01(self):
        self.assertTrue(td1 < td3)

    def test___lt__02(self):
        self.assertTrue(not td3 < td1)

    def test___ge__00(self):
        self.assertTrue(td1 >= td2)

    def test___ge__01(self):
        self.assertTrue(td3 >= td1)

    def test___ge__02(self):
        self.assertTrue(not td1 >= td3)

    def test___gt__00(self):
        self.assertTrue(not td1 > td2)

    def test___gt__01(self):
        self.assertTrue(td3 > td1)

    def test___gt__02(self):
        self.assertTrue(not td1 > td3)

    def test___bool__00(self):
        self.assertTrue(timedelta(hours=1))

    def test___bool__01(self):
        self.assertTrue(timedelta(minutes=1))

    def test___bool__02(self):
        self.assertTrue(timedelta(seconds=1))

    def test___bool__03(self):
        self.assertTrue(not td0)

    def test___str__00(self):
        self.assertEqual(str(timedelta(days=1)), "1 day, 0:00:00")

    def test___str__01(self):
        self.assertEqual(str(timedelta(days=-1)), "-1 day, 0:00:00")

    def test___str__02(self):
        self.assertEqual(str(timedelta(days=2)), "2 days, 0:00:00")

    def test___str__03(self):
        self.assertEqual(str(timedelta(days=-2)), "-2 days, 0:00:00")

    def test___str__04(self):
        self.assertEqual(str(timedelta(hours=12, minutes=58, seconds=59)), "12:58:59")

    def test___str__05(self):
        self.assertEqual(str(timedelta(hours=2, minutes=3, seconds=4)), "2:03:04")

    def test___hash__00(self):
        self.assertEqual(td0, td4)
        self.assertEqual(hash(td0), hash(td4))

    def test___hash__01(self):
        tt0 = td0 + timedelta(weeks=7)
        tt4 = td4 + timedelta(days=7 * 7)
        self.assertEqual(hash(tt0), hash(tt4))

    def test___hash__02(self):
        d = {td0: 1}
        d[td4] = 2
        self.assertEqual(len(d), 1)
        self.assertEqual(d[td0], 2)

    def test_constant00(self):
        self.assertIsInstance(timedelta.min, timedelta)
        self.assertIsInstance(timedelta.max, timedelta)
        self.assertIsInstance(timedelta.resolution, timedelta)
        self.assertTrue(timedelta.max > timedelta.min)

    def test_constant01(self):
        self.assertEqual(timedelta.min, timedelta(days=-999_999_999))

    def test_constant02(self):
        self.assertEqual(
            timedelta.max,
            timedelta(days=999_999_999, seconds=24 * 3600 - 1, microseconds=10**6 - 1),
        )

    def test_constant03(self):
        self.assertEqual(timedelta.resolution, timedelta(microseconds=1))

    def test_computation00(self):
        self.assertEqual((3 * us) * 0.5, 2 * us)

    def test_computation01(self):
        self.assertEqual((5 * us) * 0.5, 2 * us)

    def test_computation02(self):
        self.assertEqual(0.5 * (3 * us), 2 * us)

    def test_computation03(self):
        self.assertEqual(0.5 * (5 * us), 2 * us)

    def test_computation04(self):
        self.assertEqual((-3 * us) * 0.5, -2 * us)

    def test_computation05(self):
        self.assertEqual((-5 * us) * 0.5, -2 * us)

    def test_computation06(self):
        self.assertEqual((3 * us) / 2, 2 * us)

    def test_computation07(self):
        self.assertEqual((5 * us) / 2, 2 * us)

    def test_computation08(self):
        self.assertEqual((-3 * us) / 2.0, -2 * us)

    def test_computation09(self):
        self.assertEqual((-5 * us) / 2.0, -2 * us)

    def test_computation10(self):
        self.assertEqual((3 * us) / -2, -2 * us)

    def test_computation11(self):
        self.assertEqual((5 * us) / -2, -2 * us)

    def test_computation12(self):
        self.assertEqual((3 * us) / -2.0, -2 * us)

    def test_computation13(self):
        self.assertEqual((5 * us) / -2.0, -2 * us)

    def test_computation14(self):
        for i in range(-10, 10):
            # with self.subTest(i=i): not supported by Micropython
            self.assertEqual((i * us / 3) // us, round(i / 3))

    def test_computation15(self):
        for i in range(-10, 10):
            # with self.subTest(i=i): not supported by Micropython
            self.assertEqual((i * us / -3) // us, round(i / -3))

    def test_carries00(self):
        td1 = timedelta(
            days=100,
            weeks=-7,
            hours=-24 * (100 - 49),
            minutes=-3,
            seconds=3 * 60 + 1,
        )
        td2 = timedelta(seconds=1)
        self.assertEqual(td1, td2)

    def test_resolution00(self):
        self.assertIsInstance(timedelta.min, timedelta)

    def test_resolution01(self):
        self.assertIsInstance(timedelta.max, timedelta)

    def test_resolution02(self):
        self.assertIsInstance(timedelta.resolution, timedelta)

    def test_resolution03(self):
        self.assertTrue(timedelta.max > timedelta.min)

    def test_resolution04(self):
        self.assertEqual(timedelta.resolution, timedelta(microseconds=1))

    @unittest.skipIf(STDLIB, "standard timedelta has no tuple()")
    def test_tuple00(self):
        self.assertEqual(td1.tuple(), (2, 0, 0, 3, 4))

    @unittest.skipIf(STDLIB, "standard timedelta has no tuple()")
    def test_tuple01(self):
        self.assertEqual(td1h2m40s100us.tuple(), (0, 1, 2, 40, 100))


### timezone #################################################################


class Cet(tzinfo):
    # Central European Time (see https://en.wikipedia.org/wiki/Summer_time_in_Europe)

    def utcoffset(self, dt):
        h = 2 if self.isdst(dt)[0] else 1
        return timedelta(hours=h)

    def dst(self, dt):
        h = 1 if self.isdst(dt)[0] else 0
        return timedelta(hours=h)

    def tzname(self, dt):
        return "CEST" if self.isdst(dt)[0] else "CET"

    def fromutc(self, dt):
        assert dt.tzinfo is self
        isdst, fold = self.isdst(dt, utc=True)
        h = 2 if isdst else 1
        dt += timedelta(hours=h)
        dt = dt.replace(fold=fold)
        return dt

    def isdst(self, dt, utc=False):
        if dt is None:
            return False, None

        year = dt.year
        if not 2000 <= year < 2100:
            # Formulas below are valid in the range [2000; 2100)
            raise ValueError

        hour = 1 if utc else 3
        day = 31 - (5 * year // 4 + 4) % 7  # last Sunday of March
        beg = datetime(year, 3, day, hour)
        day = 31 - (5 * year // 4 + 1) % 7  # last Sunday of October
        end = datetime(year, 10, day, hour)

        dt = dt.replace(tzinfo=None)
        if utc:
            fold = 1 if end <= dt < end + timedelta(hours=1) else 0
        else:
            fold = dt.fold
        isdst = beg <= dt < end
        return isdst, fold

    def __repr__(self):
        return "Cet()"

    def __str__(self):
        return self.tzname(None)

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __hash__(self):
        return hash(repr(self))


class USTimeZone(tzinfo):
    DSTSTART = datetime(1, 3, 8, 2)
    DSTEND = datetime(1, 11, 1, 2)
    ZERO = timedelta(0)
    HOUR = timedelta(hours=1)

    def __init__(self, hours, reprname, stdname, dstname):
        self.stdoffset = timedelta(hours=hours)
        self.reprname = reprname
        self.stdname = stdname
        self.dstname = dstname

    def __repr__(self):
        return self.reprname

    def tzname(self, dt):
        if self.dst(dt):
            return self.dstname
        else:
            return self.stdname

    def utcoffset(self, dt):
        return self.stdoffset + self.dst(dt)

    def dst(self, dt):
        if dt is None or dt.tzinfo is None:
            return self.ZERO
        assert dt.tzinfo is self
        start, end = USTimeZone.us_dst_range(dt.year)
        dt = dt.replace(tzinfo=None)
        if start + self.HOUR <= dt < end - self.HOUR:
            return self.HOUR
        if end - self.HOUR <= dt < end:
            return self.ZERO if dt.fold else self.HOUR
        if start <= dt < start + self.HOUR:
            return self.HOUR if dt.fold else self.ZERO
        return self.ZERO

    def fromutc(self, dt):
        assert dt.tzinfo is self
        start, end = USTimeZone.us_dst_range(dt.year)
        start = start.replace(tzinfo=self)
        end = end.replace(tzinfo=self)
        std_time = dt + self.stdoffset
        dst_time = std_time + self.HOUR
        if end <= dst_time < end + self.HOUR:
            return std_time.replace(fold=1)
        if std_time < start or dst_time >= end:
            return std_time
        if start <= std_time < end - self.HOUR:
            return dst_time

    @staticmethod
    def us_dst_range(year):
        start = first_sunday_on_or_after(USTimeZone.DSTSTART.replace(year=year))
        end = first_sunday_on_or_after(USTimeZone.DSTEND.replace(year=year))
        return start, end

    @staticmethod
    def first_sunday_on_or_after(dt):
        days_to_go = 6 - dt.weekday()
        if days_to_go:
            dt += timedelta(days_to_go)
        return dt


class LocalTz:
    def __init__(self, tz):
        self.tz = tz
        self._old = None

    @staticmethod
    def _set(tz):
        if hasattr(mod_time, "tzset"):  # Python
            if tz:
                os.environ["TZ"] = tz
            else:
                del os.environ["TZ"]
            mod_time.tzset()
        else:
            if tz:
                os.putenv("TZ", tz)
            else:
                os.unsetenv("TZ")

    def set(self):
        self._old = os.getenv("TZ")
        LocalTz._set(self.tz)

    def unset(self):
        LocalTz._set(self._old)
        self._old = None

    def __enter__(self):
        self.set()

    def __exit__(self, typ, value, trace):
        self.unset()


tz_acdt = timezone(timedelta(hours=9.5), "ACDT")
tz_est = timezone(-timedelta(hours=5), "EST")
tz1 = timezone(timedelta(hours=-1))
tz2 = Cet()
tz3 = USTimeZone(-5, "Eastern", "EST", "EDT")


class Test1TimeZone(unittest.TestCase):
    def test___init__00(self):
        self.assertEqual(str(tz_acdt), "ACDT")
        self.assertEqual(str(tz_acdt), tz_acdt.tzname(None))

    def test___init__01(self):
        self.assertEqual(str(tz_est), "EST")
        self.assertEqual(str(tz_est), tz_est.tzname(None))

    def test___init__02(self):
        self.assertEqual(str(tz1), "UTC-01:00")
        self.assertEqual(str(tz1), tz1.tzname(None))

    def test___init__03(self):
        self.assertEqual(str(tz2), "CET")
        self.assertEqual(str(tz2), tz2.tzname(None))

    def test___init__04(self):
        offset = timedelta(hours=-24, microseconds=1)
        tz = timezone(offset)
        self.assertIsInstance(tz, timezone)

    def test___init__05(self):
        offset = timedelta(hours=24, microseconds=-1)
        tz = timezone(offset)
        self.assertIsInstance(tz, timezone)

    def test___init__06(self):
        offset = timedelta(hours=-24)
        self.assertRaises(ValueError, timezone, offset)

    def test___init__07(self):
        offset = timedelta(hours=24)
        self.assertRaises(ValueError, timezone, offset)

    def test___repr__00(self):
        self.assertEqual(tz1, eval_mod(repr(tz1)))

    def test___eq__00(self):
        self.assertEqual(timezone(timedelta(hours=1)), timezone(timedelta(hours=1)))

    def test___eq__01(self):
        self.assertNotEqual(timezone(timedelta(hours=1)), timezone(timedelta(hours=2)))

    def test___eq__02(self):
        self.assertEqual(timezone(timedelta(hours=-5)), timezone(timedelta(hours=-5), "EST"))

    def test_utcoffset00(self):
        self.assertEqual(str(tz2.utcoffset(None)), "1:00:00")

    def test_utcoffset01(self):
        self.assertEqual(str(tz2.utcoffset(datetime(2010, 3, 27, 12))), "1:00:00")

    def test_utcoffset02(self):
        self.assertEqual(str(tz2.utcoffset(datetime(2010, 3, 28, 12))), "2:00:00")

    def test_utcoffset03(self):
        self.assertEqual(str(tz2.utcoffset(datetime(2010, 10, 30, 12))), "2:00:00")

    def test_utcoffset04(self):
        self.assertEqual(str(tz2.utcoffset(datetime(2010, 10, 31, 12))), "1:00:00")

    def test_tzname00(self):
        self.assertEqual(tz2.tzname(datetime(2011, 1, 1)), "CET")

    def test_tzname01(self):
        self.assertEqual(tz2.tzname(datetime(2011, 8, 1)), "CEST")

    def test_utc00(self):
        self.assertEqual(timezone.utc.utcoffset(None), td0)

    def test_fromutc00(self):
        utc = EPOCH.replace(tzinfo=tz_acdt)
        self.assertEqual(tz_acdt.fromutc(utc), utc + 9.5 * td1h)

    def test_fromutc01(self):
        utc = EPOCH.replace(tzinfo=tz_est)
        self.assertEqual(tz_est.fromutc(utc), utc + 5 * -td1h)

    def test_fromutc02(self):
        utc = datetime(2010, 3, 28, 0, 59, 59, 999_999, tz2)
        dt = tz2.fromutc(utc)
        self.assertEqual(dt, utc + td1h)
        self.assertFalse(dt.fold)

    def test_fromutc03(self):
        utc = datetime(2010, 3, 28, 1, 0, 0, 0, tz2)
        dt = tz2.fromutc(utc)
        self.assertEqual(dt, utc + 2 * td1h)
        self.assertFalse(dt.fold)

    def test_fromutc04(self):
        utc = datetime(2010, 10, 31, 0, 59, 59, 999_999, tz2)
        dt = tz2.fromutc(utc)
        self.assertEqual(dt, utc + 2 * td1h)
        self.assertFalse(dt.fold)

    def test_fromutc05(self):
        utc = datetime(2010, 10, 31, 1, 0, 0, 0, tz2)
        dt = tz2.fromutc(utc)
        self.assertEqual(dt, utc + td1h)
        self.assertTrue(dt.fold)

    def test_fromutc06(self):
        dt1 = tz2.fromutc(datetime(2010, 10, 31, 0, 0, 0, 0, tz2))
        dt2 = tz2.fromutc(datetime(2010, 10, 31, 1, 0, 0, 0, tz2))
        self.assertEqual(dt1, dt2)
        self.assertNotEqual(dt1.fold, dt2.fold)

    def test_aware_datetime00(self):
        t = datetime(1, 1, 1)
        self.assertEqual(tz1.tzname(t), t.replace(tzinfo=tz1).tzname())

    def test_aware_datetime01(self):
        t = datetime(1, 1, 1)
        self.assertEqual(tz1.utcoffset(t), t.replace(tzinfo=tz1).utcoffset())

    def test_aware_datetime02(self):
        t = datetime(1, 1, 1)
        self.assertEqual(tz1.dst(t), t.replace(tzinfo=tz1).dst())

    def test_offset_boundaries00(self):
        td = timedelta(hours=23, minutes=59, seconds=59, microseconds=999999)
        for i in (1, -1):
            self.assertIsInstance(timezone(i * td), timezone)

    def test_offset_boundaries01(self):
        td = timedelta(hours=24)
        for i in (1, -1):
            with self.assertRaises(ValueError):
                timezone(i * td)


### date #####################################################################

d1 = date(2002, 1, 31)
d1r = "datetime.date(0, 0, 730881)"
d2 = date(1956, 1, 31)
d2d1s = (46 * 365 + len(range(1956, 2002, 4))) * 24 * 60 * 60
d3 = date(2002, 3, 1)
d4 = date(2002, 3, 2)
d5 = date(2002, 1, 31)

hour = timedelta(hours=1)
day = timedelta(days=1)
week = timedelta(weeks=1)
max_days = MAXYEAR * 365 + MAXYEAR // 4 - MAXYEAR // 100 + MAXYEAR // 400


class Test2Date(unittest.TestCase):
    def test___init__00(self):
        self.assertEqual(d1.year, 2002)
        self.assertEqual(d1.month, 1)
        self.assertEqual(d1.day, 31)

    @unittest.skipIf(STDLIB, "not supported by standard datetime")
    def test___init__01(self):
        date(0, 0, 1)

    @unittest.skipIf(STDLIB, "not supported by standard datetime")
    def test___init__02(self):
        date(0, 0, max_days)

    def test___init__03(self):
        datetime(2000, 2, 29)

    def test___init__04(self):
        datetime(2004, 2, 29)

    def test___init__05(self):
        datetime(2400, 2, 29)

    def test___init__06(self):
        self.assertRaises(ValueError, datetime, 2000, 2, 30)

    def test___init__07(self):
        self.assertRaises(ValueError, datetime, 2001, 2, 29)

    def test___init__08(self):
        self.assertRaises(ValueError, datetime, 2100, 2, 29)

    def test___init__09(self):
        self.assertRaises(ValueError, datetime, 1900, 2, 29)

    def test___init__10(self):
        self.assertRaises(ValueError, date, MINYEAR - 1, 1, 1)
        self.assertRaises(ValueError, date, MINYEAR, 0, 1)
        self.assertRaises(ValueError, date, MINYEAR, 1, 0)

    def test___init__11(self):
        self.assertRaises(ValueError, date, MAXYEAR + 1, 12, 31)
        self.assertRaises(ValueError, date, MAXYEAR, 13, 31)
        self.assertRaises(ValueError, date, MAXYEAR, 12, 32)

    def test___init__12(self):
        self.assertRaises(ValueError, date, 1, 2, 29)
        self.assertRaises(ValueError, date, 1, 4, 31)
        self.assertRaises(ValueError, date, 1, 6, 31)
        self.assertRaises(ValueError, date, 1, 9, 31)
        self.assertRaises(ValueError, date, 1, 11, 31)

    def test_fromtimestamp00(self):
        with LocalTz("UTC"):
            d = date.fromtimestamp(1012435200)
            self.assertEqual(d, d1)

    def test_fromtimestamp01(self):
        with LocalTz("UTC"):
            d = date.fromtimestamp(1012435200 + 1)
            self.assertEqual(d, d1)

    def test_fromtimestamp02(self):
        with LocalTz("UTC"):
            d = date.fromtimestamp(1012435200 - 1)
            self.assertEqual(d, d1 - timedelta(days=1))

    def test_fromtimestamp03(self):
        with LocalTz("Europe/Rome"):
            d = date.fromtimestamp(1012435200 - 3601)
            self.assertEqual(d, d1 - timedelta(days=1))

    def test_today00(self):
        tm = mod_time.localtime()[:3]
        dt = date.today()
        dd = (dt.year, dt.month, dt.day)
        self.assertEqual(tm, dd)

    def test_fromordinal00(self):
        self.assertEqual(date.fromordinal(1), date(1, 1, 1))

    def test_fromordinal01(self):
        self.assertEqual(date.fromordinal(max_days), date(MAXYEAR, 12, 31))

    def test_fromisoformat00(self):
        self.assertEqual(datetime.fromisoformat("1975-08-10"), datetime(1975, 8, 10))

    def test_year00(self):
        self.assertEqual(d1.year, 2002)

    def test_year01(self):
        self.assertEqual(d2.year, 1956)

    def test_month00(self):
        self.assertEqual(d1.month, 1)

    def test_month01(self):
        self.assertEqual(d4.month, 3)

    def test_day00(self):
        self.assertEqual(d1.day, 31)

    def test_day01(self):
        self.assertEqual(d4.day, 2)

    def test_toordinal00(self):
        self.assertEqual(date(1, 1, 1).toordinal(), 1)

    def test_toordinal01(self):
        self.assertEqual(date(MAXYEAR, 12, 31).toordinal(), max_days)

    def test_timetuple00(self):
        self.assertEqual(d1.timetuple()[:8], (2002, 1, 31, 0, 0, 0, 3, 31))

    def test_timetuple01(self):
        self.assertEqual(d3.timetuple()[:8], (2002, 3, 1, 0, 0, 0, 4, 60))

    def test_replace00(self):
        self.assertEqual(d1.replace(), d1)

    def test_replace01(self):
        self.assertEqual(d1.replace(year=2001), date(2001, 1, 31))

    def test_replace02(self):
        self.assertEqual(d1.replace(month=5), date(2002, 5, 31))

    def test_replace03(self):
        self.assertEqual(d1.replace(day=16), date(2002, 1, 16))

    def test___add__00(self):
        self.assertEqual(d4 + hour, d4)

    def test___add__01(self):
        self.assertEqual(d4 + day, date(2002, 3, 3))

    def test___add__02(self):
        self.assertEqual(d4 + week, date(2002, 3, 9))

    def test___add__03(self):
        self.assertEqual(d4 + 52 * week, date(2003, 3, 1))

    def test___add__04(self):
        self.assertEqual(d4 + -hour, date(2002, 3, 1))

    def test___add__05(self):
        self.assertEqual(d5 + -day, date(2002, 1, 30))

    def test___add__06(self):
        self.assertEqual(d4 + -week, date(2002, 2, 23))

    def test___sub__00(self):
        d = d1 - d2
        self.assertEqual(d.total_seconds(), d2d1s)

    def test___sub__01(self):
        self.assertEqual(d4 - hour, d4)

    def test___sub__02(self):
        self.assertEqual(d4 - day, date(2002, 3, 1))

    def test___sub__03(self):
        self.assertEqual(d4 - week, date(2002, 2, 23))

    def test___sub__04(self):
        self.assertEqual(d4 - 52 * week, date(2001, 3, 3))

    def test___sub__05(self):
        self.assertEqual(d4 - -hour, date(2002, 3, 3))

    def test___sub__06(self):
        self.assertEqual(d4 - -day, date(2002, 3, 3))

    def test___sub__07(self):
        self.assertEqual(d4 - -week, date(2002, 3, 9))

    def test___eq__00(self):
        self.assertEqual(d1, d5)

    def test___eq__01(self):
        self.assertFalse(d1 != d5)

    def test___eq__02(self):
        self.assertTrue(d2 != d5)

    def test___eq__03(self):
        self.assertTrue(d5 != d2)

    def test___eq__04(self):
        self.assertFalse(d2 == d5)

    def test___eq__05(self):
        self.assertFalse(d5 == d2)

    def test___eq__06(self):
        self.assertFalse(d1 == None)

    def test___eq__07(self):
        self.assertTrue(d1 != None)

    def test___le__00(self):
        self.assertTrue(d1 <= d5)

    def test___le__01(self):
        self.assertTrue(d2 <= d5)

    def test___le__02(self):
        self.assertFalse(d5 <= d2)

    def test___ge__00(self):
        self.assertTrue(d1 >= d5)

    def test___ge__01(self):
        self.assertTrue(d5 >= d2)

    def test___ge__02(self):
        self.assertFalse(d2 >= d5)

    def test___lt__00(self):
        self.assertFalse(d1 < d5)

    def test___lt__01(self):
        self.assertTrue(d2 < d5)

    def test___lt__02(self):
        self.assertFalse(d5 < d2)

    def test___gt__00(self):
        self.assertFalse(d1 > d5)

    def test___gt__01(self):
        self.assertTrue(d5 > d2)

    def test___gt__02(self):
        self.assertFalse(d2 > d5)

    def test_weekday00(self):
        for i in range(7):
            # March 4, 2002 is a Monday
            self.assertEqual(datetime(2002, 3, 4 + i).weekday(), i)
            # January 2, 1956 is a Monday
            self.assertEqual(datetime(1956, 1, 2 + i).weekday(), i)

    def test_isoweekday00(self):
        for i in range(7):
            self.assertEqual(datetime(2002, 3, 4 + i).isoweekday(), i + 1)
            self.assertEqual(datetime(1956, 1, 2 + i).isoweekday(), i + 1)

    def test_isoformat00(self):
        self.assertEqual(d1.isoformat(), "2002-01-31")

    @unittest.skipIf(STDLIB, "standard datetime differs")
    def test___repr__00(self):
        self.assertEqual(repr(d1), d1r)

    def test___repr__01(self):
        self.assertEqual(d1, eval_mod(repr(d1)))

    def test___hash__00(self):
        self.assertEqual(d1, d5)
        self.assertEqual(hash(d1), hash(d5))

    def test___hash__01(self):
        dd1 = d1 + timedelta(weeks=7)
        dd5 = d5 + timedelta(days=7 * 7)
        self.assertEqual(hash(dd1), hash(dd5))

    def test___hash__02(self):
        d = {d1: 1}
        d[d5] = 2
        self.assertEqual(len(d), 1)
        self.assertEqual(d[d1], 2)


### time #####################################################################

t1 = time(18, 45, 3, 1234)
t1r = "datetime.time(microsecond=67503001234, tzinfo=None, fold=0)"
t1f = time(18, 45, 3, 1234, fold=1)
t1fr = f"datetime.time(microsecond=67503001234, tzinfo=None, fold=1)"
t1z = time(18, 45, 3, 1234, tz1)
t1zr = f"datetime.time(microsecond=67503001234, tzinfo={repr(tz1)}, fold=0)"
t2 = time(12, 59, 59, 100)
t2z = time(12, 59, 59, 100, tz2)
t3 = time(18, 45, 3, 1234)
t3z = time(18, 45, 3, 1234, tz2)
t4 = time(18, 45, 3, 1234, fold=1)
t4z = time(18, 45, 3, 1234, tz2, fold=1)
t5z = time(20, 45, 3, 1234, tz2)


class Test3Time(unittest.TestCase):
    def test___init__00(self):
        t = time()
        self.assertEqual(t.hour, 0)
        self.assertEqual(t.minute, 0)
        self.assertEqual(t.second, 0)
        self.assertEqual(t.microsecond, 0)
        self.assertEqual(t.tzinfo, None)
        self.assertEqual(t.fold, 0)

    def test___init__01(self):
        t = time(12)
        self.assertEqual(t.hour, 12)
        self.assertEqual(t.minute, 0)
        self.assertEqual(t.second, 0)
        self.assertEqual(t.microsecond, 0)
        self.assertEqual(t.tzinfo, None)
        self.assertEqual(t.fold, 0)

    def test___init__02(self):
        self.assertEqual(t1z.hour, 18)
        self.assertEqual(t1z.minute, 45)
        self.assertEqual(t1z.second, 3)
        self.assertEqual(t1z.microsecond, 1234)
        self.assertEqual(t1z.tzinfo, tz1)
        self.assertEqual(t1z.fold, 0)

    def test___init__03(self):
        t = time(microsecond=1, fold=1)
        self.assertEqual(t.fold, 1)

    @unittest.skipIf(STDLIB, "not supported by standard datetime")
    def test___init__04(self):
        time(microsecond=24 * 60 * 60 * 1_000_000 - 1)

    def test___init__05(self):
        self.assertRaises(ValueError, time, -1, 0, 0, 0)
        self.assertRaises(ValueError, time, 0, -1, 0, 0)
        self.assertRaises(ValueError, time, 0, 0, -1, 0)
        self.assertRaises(ValueError, time, 0, 0, 0, -1)
        self.assertRaises(ValueError, time, 0, 0, 0, 0, fold=-1)

    def test___init__06(self):
        self.assertRaises(ValueError, time, 24, 0, 0, 0)
        self.assertRaises(ValueError, time, 0, 60, 0, 0)
        self.assertRaises(ValueError, time, 0, 0, 60, 0)
        self.assertRaises(ValueError, time, 0, 0, 0, 0, fold=2)

    @unittest.skipIf(STDLIB, "not supported by standard datetime")
    def test___init__07(self):
        self.assertRaises(ValueError, time, microsecond=24 * 60 * 60 * 1_000_000)

    def test_fromisoformat00(self):
        self.assertEqual(time.fromisoformat("01"), time(1))

    def test_fromisoformat01(self):
        self.assertEqual(time.fromisoformat("13:30"), time(13, 30))

    def test_fromisoformat02(self):
        self.assertEqual(time.fromisoformat("23:30:12"), time(23, 30, 12))

    def test_fromisoformat03(self):
        self.assertEqual(str(time.fromisoformat("11:03:04+01:00")), "11:03:04+01:00")

    def test_hour00(self):
        self.assertEqual(t1.hour, 18)

    def test_hour01(self):
        self.assertEqual(t2z.hour, 12)

    def test_minute00(self):
        self.assertEqual(t1.minute, 45)

    def test_minute01(self):
        self.assertEqual(t2z.minute, 59)

    def test_second00(self):
        self.assertEqual(t1.second, 3)

    def test_second01(self):
        self.assertEqual(t2z.second, 59)

    def test_microsecond00(self):
        self.assertEqual(t1.microsecond, 1234)

    def test_microsecond01(self):
        self.assertEqual(t2z.microsecond, 100)

    def test_tzinfo00(self):
        self.assertEqual(t1.tzinfo, None)

    def test_tzinfo01(self):
        self.assertEqual(t2z.tzinfo, tz2)

    def test_fold00(self):
        self.assertEqual(t1.fold, 0)

    def test_replace00(self):
        self.assertEqual(t2z.replace(), t2z)

    def test_replace01(self):
        self.assertEqual(t2z.replace(hour=20), time(20, 59, 59, 100, tz2))

    def test_replace02(self):
        self.assertEqual(t2z.replace(minute=4), time(12, 4, 59, 100, tz2))

    def test_replace03(self):
        self.assertEqual(t2z.replace(second=16), time(12, 59, 16, 100, tz2))

    def test_replace04(self):
        self.assertEqual(t2z.replace(microsecond=99), time(12, 59, 59, 99, tz2))

    def test_replace05(self):
        self.assertEqual(t2z.replace(tzinfo=tz1), time(12, 59, 59, 100, tz1))

    def test_isoformat00(self):
        self.assertEqual(t1.isoformat(), "18:45:03.001234")

    def test_isoformat01(self):
        self.assertEqual(t1z.isoformat(), "18:45:03.001234-01:00")

    def test_isoformat02(self):
        self.assertEqual(t2z.isoformat(), "12:59:59.000100+01:00")

    @unittest.skipIf(STDLIB, "standard datetime differs")
    def test___repr__00(self):
        self.assertEqual(repr(t1), t1r)

    @unittest.skipIf(STDLIB, "standard datetime differs")
    def test___repr__01(self):
        self.assertEqual(repr(t1f), t1fr)

    @unittest.skipIf(STDLIB, "standard datetime differs")
    def test___repr__02(self):
        self.assertEqual(repr(t1z), t1zr)

    def test___repr__03(self):
        self.assertEqual(t1, eval_mod(repr(t1)))

    def test___repr__04(self):
        self.assertEqual(t1z, eval_mod(repr(t1z)))

    def test___repr__05(self):
        self.assertEqual(t4, eval_mod(repr(t4)))

    def test___repr__06(self):
        dt = eval_mod(repr(t4z))
        self.assertEqual(t4z, eval_mod(repr(t4z)))

    def test___bool__00(self):
        self.assertTrue(t1)

    def test___bool__01(self):
        self.assertTrue(t1z)

    def test___bool__02(self):
        self.assertTrue(time())

    def test___eq__00(self):
        self.assertEqual(t1, t1)

    def test___eq__01(self):
        self.assertEqual(t1z, t1z)

    def test___eq__02(self):
        self.assertNotEqual(t1, t1z)

    def test___eq__03(self):
        self.assertNotEqual(t1z, t2z)

    def test___eq__04(self):
        self.assertEqual(t1z, t5z)

    def test___eq__05(self):
        self.assertEqual(t1, t1f)

    def test___lt__00(self):
        self.assertTrue(t2 < t1)

    def test___lt__01(self):
        self.assertTrue(t2z < t1z)

    def test___lt__02(self):
        self.assertRaises(TypeError, t1.__lt__, t1z)

    def test___le__00(self):
        self.assertTrue(t3 <= t1)

    def test___le__01(self):
        self.assertTrue(t1z <= t5z)

    def test___le__02(self):
        self.assertRaises(TypeError, t1.__le__, t1z)

    def test___ge__00(self):
        self.assertTrue(t1 >= t3)

    def test___ge__01(self):
        self.assertTrue(t5z >= t1z)

    def test___ge__02(self):
        self.assertRaises(TypeError, t1.__ge__, t1z)

    def test___gt__00(self):
        self.assertTrue(t1 > t2)

    def test___gt__01(self):
        self.assertTrue(t1z > t2z)

    def test___gt__02(self):
        self.assertRaises(TypeError, t1.__gt__, t1z)

    def test___hash__00(self):
        self.assertEqual(t1, t3)
        self.assertEqual(hash(t1), hash(t3))

    def test___hash__01(self):
        d = {t1: 1}
        d[t3] = 3
        self.assertEqual(len(d), 1)
        self.assertEqual(d[t1], 3)

    def test___hash__02(self):
        self.assertNotEqual(t1, t1z)
        self.assertNotEqual(hash(t1), hash(t1z))

    def test___hash__03(self):
        self.assertNotEqual(t1z, t3z)
        self.assertNotEqual(hash(t1z), hash(t3z))

    def test___hash__04(self):
        tf = t1.replace(fold=1)
        self.assertEqual(t1, tf)
        self.assertEqual(hash(t1), hash(tf))

    def test_utcoffset00(self):
        self.assertEqual(t1.utcoffset(), None)

    def test_utcoffset01(self):
        self.assertEqual(t1z.utcoffset(), timedelta(hours=-1))

    def test_utcoffset02(self):
        self.assertEqual(t2z.utcoffset(), timedelta(hours=1))

    def test_dst00(self):
        self.assertEqual(t1.dst(), None)

    def test_dst01(self):
        self.assertEqual(t1z.dst(), None)

    def test_dst02(self):
        self.assertEqual(t2z.dst(), td0)

    def test_tzname00(self):
        self.assertEqual(t1.tzname(), None)

    def test_tzname01(self):
        self.assertEqual(t1z.tzname(), "UTC-01:00")

    def test_tzname02(self):
        self.assertEqual(t2z.tzname(), "CET")

    def test_constant00(self):
        self.assertIsInstance(timedelta.resolution, timedelta)
        self.assertTrue(timedelta.max > timedelta.min)

    def test_constant01(self):
        self.assertEqual(time.min, time(0))

    def test_constant02(self):
        self.assertEqual(time.max, time(23, 59, 59, 999_999))

    def test_constant03(self):
        self.assertEqual(time.resolution, timedelta(microseconds=1))


### datetime #################################################################

dt1 = datetime(2002, 1, 31)
dt1z1 = datetime(2002, 1, 31, tzinfo=tz1)
dt1z2 = datetime(2002, 1, 31, tzinfo=tz2)
dt2 = datetime(1956, 1, 31)
dt3 = datetime(2002, 3, 1, 12, 59, 59, 100, tz2)
dt4 = datetime(2002, 3, 2, 17, 6)
dt5 = datetime(2002, 1, 31)
dt5z2 = datetime(2002, 1, 31, tzinfo=tz2)

dt1r = "datetime.datetime(2002, 1, 31, 0, 0, 0, 0, None, fold=0)"
dt3r = "datetime.datetime(2002, 3, 1, 12, 59, 59, 100, Cet(), fold=0)"
dt4r = "datetime.datetime(2002, 3, 2, 17, 6, 0, 0, None, fold=0)"

d1t1 = datetime(2002, 1, 31, 18, 45, 3, 1234)
d1t1f = datetime(2002, 1, 31, 18, 45, 3, 1234, fold=1)
d1t1z = datetime(2002, 1, 31, 18, 45, 3, 1234, tz1)

dt27tz2 = datetime(2010, 3, 27, 12, tzinfo=tz2)  # last CET day
dt28tz2 = datetime(2010, 3, 28, 12, tzinfo=tz2)  # first CEST day
dt30tz2 = datetime(2010, 10, 30, 12, tzinfo=tz2)  # last CEST day
dt31tz2 = datetime(2010, 10, 31, 12, tzinfo=tz2)  # first CET day


# Tests where datetime depens on date and time
class Test4DateTime(unittest.TestCase):
    def test_combine00(self):
        dt = datetime.combine(d1, t1)
        self.assertEqual(dt, d1t1)

    def test_combine01(self):
        dt = datetime.combine(d1, t1)
        self.assertEqual(dt.date(), d1)

    def test_combine02(self):
        dt1 = datetime.combine(d1, t1)
        dt2 = datetime.combine(dt1, t1)
        self.assertEqual(dt1, dt2)

    def test_combine03(self):
        dt = datetime.combine(d1, t1)
        self.assertEqual(dt.time(), t1)

    def test_combine04(self):
        dt = datetime.combine(d1, t1, tz1)
        self.assertEqual(dt, d1t1z)

    def test_combine05(self):
        dt = datetime.combine(d1, t1z)
        self.assertEqual(dt, d1t1z)

    def test_combine06(self):
        dt = datetime.combine(d1, t1f)
        self.assertEqual(dt, d1t1f)

    def test_date00(self):
        self.assertEqual(d1t1.date(), d1)

    def test_time00(self):
        self.assertEqual(d1t1.time(), t1)

    def test_time01(self):
        self.assertNotEqual(d1t1z.time(), t1z)

    def test_timetz00(self):
        self.assertEqual(d1t1.timetz(), t1)

    def test_timetz01(self):
        self.assertEqual(d1t1z.timetz(), t1z)

    def test_timetz02(self):
        self.assertEqual(d1t1f.timetz(), t1f)


# Tests where datetime is independent from date and time
class Test5DateTime(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        for k in ("date", "time"):
            del mod_datetime.__dict__[k]

    def test___init__00(self):
        d = datetime(2002, 3, 1, 12, 0, fold=1)
        self.assertEqual(d.year, 2002)
        self.assertEqual(d.month, 3)
        self.assertEqual(d.day, 1)
        self.assertEqual(d.hour, 12)
        self.assertEqual(d.minute, 0)
        self.assertEqual(d.second, 0)
        self.assertEqual(d.microsecond, 0)
        self.assertEqual(d.tzinfo, None)
        self.assertEqual(d.fold, 1)

    def test___init__01(self):
        self.assertEqual(dt3.year, 2002)
        self.assertEqual(dt3.month, 3)
        self.assertEqual(dt3.day, 1)
        self.assertEqual(dt3.hour, 12)
        self.assertEqual(dt3.minute, 59)
        self.assertEqual(dt3.second, 59)
        self.assertEqual(dt3.microsecond, 100)
        self.assertEqual(dt3.tzinfo, tz2)
        self.assertEqual(dt3.fold, 0)

    def test___init__02(self):
        datetime(MINYEAR, 1, 1)

    def test___init__03(self):
        datetime(MAXYEAR, 12, 31)

    def test___init__04(self):
        self.assertRaises(ValueError, datetime, MINYEAR - 1, 1, 1)

    def test___init__05(self):
        self.assertRaises(ValueError, datetime, MAXYEAR + 1, 1, 1)

    def test___init__06(self):
        self.assertRaises(ValueError, datetime, 2000, 0, 1)

    def test___init__07(self):
        datetime(2000, 2, 29)

    def test___init__08(self):
        datetime(2004, 2, 29)

    def test___init__09(self):
        datetime(2400, 2, 29)

    def test___init__10(self):
        self.assertRaises(ValueError, datetime, 2000, 2, 30)

    def test___init__11(self):
        self.assertRaises(ValueError, datetime, 2001, 2, 29)

    def test___init__12(self):
        self.assertRaises(ValueError, datetime, 2100, 2, 29)

    def test___init__13(self):
        self.assertRaises(ValueError, datetime, 1900, 2, 29)

    def test___init__14(self):
        self.assertRaises(ValueError, datetime, 2000, 1, 0)

    def test___init__15(self):
        self.assertRaises(ValueError, datetime, 2000, 1, 32)

    def test___init__16(self):
        self.assertRaises(ValueError, datetime, 2000, 1, 31, -1)

    def test___init__17(self):
        self.assertRaises(ValueError, datetime, 2000, 1, 31, 24)

    def test___init__18(self):
        self.assertRaises(ValueError, datetime, 2000, 1, 31, 23, -1)

    def test___init__19(self):
        self.assertRaises(ValueError, datetime, 2000, 1, 31, 23, 60)

    def test___init__20(self):
        self.assertRaises(ValueError, datetime, 2000, 1, 31, 23, 59, -1)

    def test___init__21(self):
        self.assertRaises(ValueError, datetime, 2000, 1, 31, 23, 59, 60)

    def test___init__22(self):
        self.assertEqual(dt1, eval_mod(dt1r))

    def test___init__23(self):
        self.assertEqual(dt3, eval_mod(dt3r))

    def test___init__24(self):
        self.assertEqual(dt4, eval_mod(dt4r))

    def test_fromtimestamp00(self):
        with LocalTz("Europe/Rome"):
            ts = 1012499103.001234
            if LOCALTZ:
                dt = datetime.fromtimestamp(ts)
                self.assertEqual(dt, d1t1)
            else:
                self.assertRaises(NotImplementedError, datetime.fromtimestamp, ts)

    def test_fromtimestamp01(self):
        ts = 1012506303.001234
        self.assertEqual(datetime.fromtimestamp(ts, tz1), d1t1z)

    def test_fromtimestamp02(self):
        ts = 1269687600
        self.assertEqual(datetime.fromtimestamp(ts, tz2), dt27tz2)

    def test_fromtimestamp03(self):
        ts = 1269770400
        self.assertEqual(datetime.fromtimestamp(ts, tz2), dt28tz2)

    def test_fromtimestamp04(self):
        with LocalTz("Europe/Rome"):
            dt = datetime(2010, 10, 31, 0, 30, tzinfo=timezone.utc)
            ts = (dt - EPOCH).total_seconds()
            dt = dt.replace(tzinfo=None) + 2 * td1h
            if LOCALTZ:
                ds = datetime.fromtimestamp(ts)
                self.assertEqual(ds, dt)
                self.assertFalse(ds.fold)
            else:
                self.assertRaises(NotImplementedError, datetime.fromtimestamp, ts)

    def test_fromtimestamp05(self):
        with LocalTz("Europe/Rome"):
            dt = datetime(2010, 10, 31, 1, 30, tzinfo=timezone.utc)
            ts = (dt - EPOCH).total_seconds()
            dt = dt.replace(tzinfo=None) + 1 * td1h
            if LOCALTZ:
                ds = datetime.fromtimestamp(ts)
                self.assertEqual(ds, dt)
                self.assertTrue(ds.fold)
            else:
                self.assertRaises(NotImplementedError, datetime.fromtimestamp, ts)

    def test_fromtimestamp06(self):
        with LocalTz("US/Eastern"):
            dt = datetime(2020, 11, 1, 5, 30, tzinfo=timezone.utc)
            ts = (dt - EPOCH).total_seconds()
            dt = dt.replace(tzinfo=None) - 4 * td1h
            if LOCALTZ:
                ds = datetime.fromtimestamp(ts)
                self.assertEqual(ds, dt)
            else:
                self.assertRaises(NotImplementedError, datetime.fromtimestamp, ts)

    def test_fromtimestamp07(self):
        with LocalTz("US/Eastern"):
            dt = datetime(2020, 11, 1, 7, 30, tzinfo=timezone.utc)
            ts = (dt - EPOCH).total_seconds()
            dt = dt.replace(tzinfo=None) - 5 * td1h
            if LOCALTZ:
                ds = datetime.fromtimestamp(ts)
                self.assertEqual(ds, dt)
            else:
                self.assertRaises(NotImplementedError, datetime.fromtimestamp, ts)

    @unittest.skipIf(not LOCALTZ, "naive datetime not supported")
    def test_now00(self):
        tm = datetime(*mod_time.localtime()[:6])
        dt = datetime.now()
        self.assertAlmostEqual(tm, dt, delta=timedelta(seconds=1))

    def test_now01(self):
        tm = datetime(*mod_time.gmtime()[:6], tzinfo=tz2)
        tm += tz2.utcoffset(tm)
        dt = datetime.now(tz2)
        self.assertAlmostEqual(tm, dt, delta=timedelta(seconds=1))

    def test_fromordinal00(self):
        self.assertEqual(datetime.fromordinal(1), datetime(1, 1, 1))

    def test_fromordinal01(self):
        self.assertEqual(datetime.fromordinal(max_days), datetime(MAXYEAR, 12, 31))

    def test_fromisoformat00(self):
        self.assertEqual(datetime.fromisoformat("1975-08-10"), datetime(1975, 8, 10))

    def test_fromisoformat01(self):
        self.assertEqual(datetime.fromisoformat("1975-08-10 23"), datetime(1975, 8, 10, 23))

    def test_fromisoformat02(self):
        self.assertEqual(datetime.fromisoformat("1975-08-10 23:30"), datetime(1975, 8, 10, 23, 30))

    def test_fromisoformat03(self):
        self.assertEqual(
            datetime.fromisoformat("1975-08-10 23:30:12"), datetime(1975, 8, 10, 23, 30, 12)
        )

    def test_fromisoformat04(self):
        self.assertEqual(
            str(datetime.fromisoformat("1975-08-10 23:30:12+01:00")), "1975-08-10 23:30:12+01:00"
        )

    def test_year00(self):
        self.assertEqual(dt1.year, 2002)

    def test_year01(self):
        self.assertEqual(dt2.year, 1956)

    def test_month00(self):
        self.assertEqual(dt1.month, 1)

    def test_month01(self):
        self.assertEqual(dt3.month, 3)

    def test_day00(self):
        self.assertEqual(dt1.day, 31)

    def test_day01(self):
        self.assertEqual(dt4.day, 2)

    def test_hour00(self):
        self.assertEqual(dt1.hour, 0)

    def test_hour01(self):
        self.assertEqual(dt3.hour, 12)

    def test_minute00(self):
        self.assertEqual(dt1.minute, 0)

    def test_minute01(self):
        self.assertEqual(dt3.minute, 59)

    def test_second00(self):
        self.assertEqual(dt1.second, 0)

    def test_second01(self):
        self.assertEqual(dt3.second, 59)

    def test_microsecond00(self):
        self.assertEqual(dt1.microsecond, 0)

    def test_microsecond01(self):
        self.assertEqual(dt3.microsecond, 100)

    def test_tzinfo00(self):
        self.assertEqual(dt1.tzinfo, None)

    def test_tzinfo01(self):
        self.assertEqual(dt3.tzinfo, tz2)

    def test_fold00(self):
        self.assertEqual(dt1.fold, 0)

    def test___add__00(self):
        self.assertEqual(dt4 + hour, datetime(2002, 3, 2, 18, 6))

    def test___add__01(self):
        self.assertEqual(hour + dt4, datetime(2002, 3, 2, 18, 6))

    def test___add__02(self):
        self.assertEqual(dt4 + 10 * hour, datetime(2002, 3, 3, 3, 6))

    def test___add__03(self):
        self.assertEqual(dt4 + day, datetime(2002, 3, 3, 17, 6))

    def test___add__04(self):
        self.assertEqual(dt4 + week, datetime(2002, 3, 9, 17, 6))

    def test___add__05(self):
        self.assertEqual(dt4 + 52 * week, datetime(2003, 3, 1, 17, 6))

    def test___add__06(self):
        self.assertEqual(dt4 + (week + day + hour), datetime(2002, 3, 10, 18, 6))

    def test___add__07(self):
        self.assertEqual(dt5 + -day, datetime(2002, 1, 30))

    def test___add__08(self):
        self.assertEqual(-hour + dt4, datetime(2002, 3, 2, 16, 6))

    def test___sub__00(self):
        d = dt1 - dt2
        self.assertEqual(d.total_seconds(), d2d1s)

    def test___sub__01(self):
        self.assertEqual(dt4 - hour, datetime(2002, 3, 2, 16, 6))

    def test___sub__02(self):
        self.assertEqual(dt4 - hour, dt4 + -hour)

    def test___sub__03(self):
        self.assertEqual(dt4 - 20 * hour, datetime(2002, 3, 1, 21, 6))

    def test___sub__04(self):
        self.assertEqual(dt4 - day, datetime(2002, 3, 1, 17, 6))

    def test___sub__05(self):
        self.assertEqual(dt4 - week, datetime(2002, 2, 23, 17, 6))

    def test___sub__06(self):
        self.assertEqual(dt4 - 52 * week, datetime(2001, 3, 3, 17, 6))

    def test___sub__07(self):
        self.assertEqual(dt4 - (week + day + hour), datetime(2002, 2, 22, 16, 6))

    def test_computation00(self):
        self.assertEqual((dt4 + week) - dt4, week)

    def test_computation01(self):
        self.assertEqual((dt4 + day) - dt4, day)

    def test_computation02(self):
        self.assertEqual((dt4 + hour) - dt4, hour)

    def test_computation03(self):
        self.assertEqual(dt4 - (dt4 + week), -week)

    def test_computation04(self):
        self.assertEqual(dt4 - (dt4 + day), -day)

    def test_computation05(self):
        self.assertEqual(dt4 - (dt4 + hour), -hour)

    def test_computation06(self):
        self.assertEqual(dt4 - (dt4 - week), week)

    def test_computation07(self):
        self.assertEqual(dt4 - (dt4 - day), day)

    def test_computation08(self):
        self.assertEqual(dt4 - (dt4 - hour), hour)

    def test_computation09(self):
        self.assertEqual(dt4 + (week + day + hour), (((dt4 + week) + day) + hour))

    def test_computation10(self):
        self.assertEqual(dt4 - (week + day + hour), (((dt4 - week) - day) - hour))

    def test___eq__00(self):
        self.assertEqual(dt1, dt5)

    def test___eq__01(self):
        self.assertFalse(dt1 != dt5)

    def test___eq__02(self):
        self.assertTrue(dt2 != dt5)

    def test___eq__03(self):
        self.assertTrue(dt5 != dt2)

    def test___eq__04(self):
        self.assertFalse(dt2 == dt5)

    def test___eq__05(self):
        self.assertFalse(dt5 == dt2)

    def test___eq__06(self):
        self.assertFalse(dt1 == dt1z1)

    def test___eq__07(self):
        self.assertFalse(dt1z1 == dt1z2)

    def test___eq__08(self):
        self.assertTrue(dt1z2 == dt5z2)

    def test___le__00(self):
        self.assertTrue(dt1 <= dt5)

    def test___le__01(self):
        self.assertTrue(dt2 <= dt5)

    def test___le__02(self):
        self.assertFalse(dt5 <= dt2)

    def test___le__03(self):
        self.assertFalse(dt1z1 <= dt1z2)

    def test___le__04(self):
        self.assertTrue(dt1z2 <= dt5z2)

    def test___le__05(self):
        self.assertRaises(TypeError, dt1.__le__, dt1z1)

    def test___ge__00(self):
        self.assertTrue(dt1 >= dt5)

    def test___ge__01(self):
        self.assertTrue(dt5 >= dt2)

    def test___ge__02(self):
        self.assertFalse(dt2 >= dt5)

    def test___ge__03(self):
        self.assertTrue(dt1z1 >= dt1z2)

    def test___ge__04(self):
        self.assertTrue(dt1z2 >= dt5z2)

    def test___ge__05(self):
        self.assertRaises(TypeError, dt1.__ge__, dt1z1)

    def test___lt__00(self):
        self.assertFalse(dt1 < dt5)

    def test___lt__01(self):
        self.assertTrue(dt2 < dt5)

    def test___lt__02(self):
        self.assertFalse(dt5 < dt2)

    def test___lt__03(self):
        self.assertFalse(dt1z1 < dt1z2)

    def test___lt__04(self):
        self.assertFalse(dt1z2 < dt5z2)

    def test___lt__05(self):
        self.assertRaises(TypeError, dt1.__lt__, dt1z1)

    def test___gt__00(self):
        self.assertFalse(dt1 > dt5)

    def test___gt__01(self):
        self.assertTrue(dt5 > dt2)

    def test___gt__02(self):
        self.assertFalse(dt2 > dt5)

    def test___gt__03(self):
        self.assertTrue(dt1z1 > dt1z2)

    def test___gt__04(self):
        self.assertFalse(dt1z2 > dt5z2)

    def test___gt__05(self):
        self.assertRaises(TypeError, dt1.__gt__, dt1z1)

    def test_replace00(self):
        self.assertEqual(dt3.replace(), dt3)

    def test_replace01(self):
        self.assertEqual(dt3.replace(year=2001), datetime(2001, 3, 1, 12, 59, 59, 100, tz2))

    def test_replace02(self):
        self.assertEqual(dt3.replace(month=4), datetime(2002, 4, 1, 12, 59, 59, 100, tz2))

    def test_replace03(self):
        self.assertEqual(dt3.replace(day=16), datetime(2002, 3, 16, 12, 59, 59, 100, tz2))

    def test_replace04(self):
        self.assertEqual(dt3.replace(hour=13), datetime(2002, 3, 1, 13, 59, 59, 100, tz2))

    def test_replace05(self):
        self.assertEqual(dt3.replace(minute=0), datetime(2002, 3, 1, 12, 0, 59, 100, tz2))

    def test_replace06(self):
        self.assertEqual(dt3.replace(second=1), datetime(2002, 3, 1, 12, 59, 1, 100, tz2))

    def test_replace07(self):
        self.assertEqual(dt3.replace(microsecond=99), datetime(2002, 3, 1, 12, 59, 59, 99, tz2))

    def test_replace08(self):
        self.assertEqual(dt3.replace(tzinfo=tz1), datetime(2002, 3, 1, 12, 59, 59, 100, tz1))

    def test_replace09(self):
        self.assertRaises(ValueError, datetime(2000, 2, 29).replace, year=2001)

    def test_astimezone00(self):
        dt = datetime(2002, 3, 1, 11, 59, 59, 100, timezone.utc)
        self.assertEqual(dt3.astimezone(timezone.utc), dt)

    def test_astimezone01(self):
        self.assertIs(dt1z1.astimezone(tz1), dt1z1)

    def test_astimezone02(self):
        dt = datetime(2002, 1, 31, 2, 0, tzinfo=tz2)
        self.assertEqual(dt1z1.astimezone(tz2), dt)

    def test_astimezone03(self):
        dt = datetime(2002, 1, 31, 10, 30, tzinfo=tz_acdt)
        self.assertEqual(dt1z1.astimezone(tz_acdt), dt)

    def test_astimezone04(self):
        with LocalTz("Europe/Rome"):
            dt1 = dt27tz2
            dt2 = dt1.replace(tzinfo=None)
            if LOCALTZ:
                self.assertEqual(dt1, dt2.astimezone(tz2))
            else:
                self.assertRaises(NotImplementedError, dt2.astimezone, tz2)

    def test_astimezone05(self):
        with LocalTz("Europe/Rome"):
            dt1 = dt28tz2
            dt2 = dt1.replace(tzinfo=None)
            if LOCALTZ:
                self.assertEqual(dt1, dt2.astimezone(tz2))
            else:
                self.assertRaises(NotImplementedError, dt2.astimezone, tz2)

    def test_astimezone06(self):
        with LocalTz("Europe/Rome"):
            dt1 = dt30tz2
            dt2 = dt1.replace(tzinfo=None)
            if LOCALTZ:
                self.assertEqual(dt1, dt2.astimezone(tz2))
            else:
                self.assertRaises(NotImplementedError, dt2.astimezone, tz2)

    def test_astimezone07(self):
        with LocalTz("Europe/Rome"):
            dt1 = dt31tz2
            dt2 = dt1.replace(tzinfo=None)
            if LOCALTZ:
                self.assertEqual(dt1, dt2.astimezone(tz2))
            else:
                self.assertRaises(NotImplementedError, dt2.astimezone, tz2)

    def test_astimezone08(self):
        with LocalTz("Europe/Rome"):
            dt1 = dt3
            dt2 = dt1.replace(tzinfo=None)
            if LOCALTZ:
                self.assertEqual(dt1, dt2.astimezone(tz2))
            else:
                self.assertRaises(NotImplementedError, dt2.astimezone, tz2)

    def test_utcoffset00(self):
        self.assertEqual(dt1.utcoffset(), None)

    def test_utcoffset01(self):
        self.assertEqual(dt27tz2.utcoffset(), timedelta(hours=1))

    def test_utcoffset02(self):
        self.assertEqual(dt28tz2.utcoffset(), timedelta(hours=2))

    def test_utcoffset03(self):
        self.assertEqual(dt30tz2.utcoffset(), timedelta(hours=2))

    def test_utcoffset04(self):
        self.assertEqual(dt31tz2.utcoffset(), timedelta(hours=1))

    def test_dst00(self):
        self.assertEqual(dt1.dst(), None)

    def test_dst01(self):
        self.assertEqual(dt27tz2.dst(), timedelta(hours=0))

    def test_dst02(self):
        self.assertEqual(dt28tz2.dst(), timedelta(hours=1))

    def test_tzname00(self):
        self.assertEqual(dt1.tzname(), None)

    def test_tzname01(self):
        self.assertEqual(dt27tz2.tzname(), "CET")

    def test_tzname02(self):
        self.assertEqual(dt28tz2.tzname(), "CEST")

    def test_timetuple00(self):
        with LocalTz("Europe/Rome"):
            self.assertEqual(dt1.timetuple()[:8], (2002, 1, 31, 0, 0, 0, 3, 31))

    @unittest.skip("broken when running with non-UTC timezone")
    def test_timetuple01(self):
        self.assertEqual(dt27tz2.timetuple()[:8], (2010, 3, 27, 12, 0, 0, 5, 86))

    @unittest.skip("broken when running with non-UTC timezone")
    def test_timetuple02(self):
        self.assertEqual(dt28tz2.timetuple()[:8], (2010, 3, 28, 12, 0, 0, 6, 87))

    def test_timetuple03(self):
        with LocalTz("Europe/Rome"):
            self.assertEqual(
                dt27tz2.replace(tzinfo=None).timetuple()[:8], (2010, 3, 27, 12, 0, 0, 5, 86)
            )

    def test_timetuple04(self):
        self.assertEqual(
            dt28tz2.replace(tzinfo=None).timetuple()[:8], (2010, 3, 28, 12, 0, 0, 6, 87)
        )

    def test_toordinal00(self):
        self.assertEqual(datetime(1, 1, 1).toordinal(), 1)

    def test_toordinal01(self):
        self.assertEqual(datetime(1, 12, 31).toordinal(), 365)

    def test_toordinal02(self):
        self.assertEqual(datetime(2, 1, 1).toordinal(), 366)

    def test_toordinal03(self):
        # https://www.timeanddate.com/date/dateadded.html?d1=1&m1=1&y1=1&type=add&ad=730882
        self.assertEqual(dt1.toordinal(), 730_882 - 1)

    def test_toordinal04(self):
        # https://www.timeanddate.com/date/dateadded.html?d1=1&m1=1&y1=1&type=add&ad=730911
        self.assertEqual(dt3.toordinal(), 730_911 - 1)

    def test_weekday00(self):
        self.assertEqual(dt1.weekday(), d1.weekday())

    def test_timestamp00(self):
        with LocalTz("Europe/Rome"):
            if LOCALTZ:
                self.assertEqual(d1t1.timestamp(), 1012499103.001234)
            else:
                self.assertRaises(NotImplementedError, d1t1.timestamp)

    def test_timestamp01(self):
        self.assertEqual(d1t1z.timestamp(), 1012506303.001234)

    def test_timestamp02(self):
        with LocalTz("Europe/Rome"):
            dt = datetime(2010, 3, 28, 2, 30)  # doens't exist
            if LOCALTZ:
                self.assertEqual(dt.timestamp(), 1269739800.0)
            else:
                self.assertRaises(NotImplementedError, dt.timestamp)

    def test_timestamp03(self):
        with LocalTz("Europe/Rome"):
            dt = datetime(2010, 8, 10, 2, 30)
            if LOCALTZ:
                self.assertEqual(dt.timestamp(), 1281400200.0)
            else:
                self.assertRaises(NotImplementedError, dt.timestamp)

    def test_timestamp04(self):
        with LocalTz("Europe/Rome"):
            dt = datetime(2010, 10, 31, 2, 30, fold=0)
            if LOCALTZ:
                self.assertEqual(dt.timestamp(), 1288485000.0)
            else:
                self.assertRaises(NotImplementedError, dt.timestamp)

    def test_timestamp05(self):
        with LocalTz("Europe/Rome"):
            dt = datetime(2010, 10, 31, 2, 30, fold=1)
            if LOCALTZ:
                self.assertEqual(dt.timestamp(), 1288488600.0)
            else:
                self.assertRaises(NotImplementedError, dt.timestamp)

    def test_timestamp06(self):
        with LocalTz("US/Eastern"):
            dt = datetime(2020, 3, 8, 2, 30)  # doens't exist
            if LOCALTZ:
                self.assertEqual(dt.timestamp(), 1583652600.0)
            else:
                self.assertRaises(NotImplementedError, dt.timestamp)

    def test_timestamp07(self):
        with LocalTz("US/Eastern"):
            dt = datetime(2020, 8, 10, 2, 30)
            if LOCALTZ:
                self.assertEqual(dt.timestamp(), 1597041000.0)
            else:
                self.assertRaises(NotImplementedError, dt.timestamp)

    def test_timestamp08(self):
        with LocalTz("US/Eastern"):
            dt = datetime(2020, 11, 1, 2, 30, fold=0)
            if LOCALTZ:
                self.assertEqual(dt.timestamp(), 1604215800.0)
            else:
                self.assertRaises(NotImplementedError, dt.timestamp)

    def test_timestamp09(self):
        with LocalTz("US/Eastern"):
            dt = datetime(2020, 11, 1, 2, 30, fold=1)
            if LOCALTZ:
                self.assertEqual(dt.timestamp(), 1604215800.0)
            else:
                self.assertRaises(NotImplementedError, dt.timestamp)

    def test_isoweekday00(self):
        self.assertEqual(dt1.isoweekday(), d1.isoweekday())

    def test_isoformat00(self):
        self.assertEqual(dt3.isoformat(), "2002-03-01T12:59:59.000100+01:00")

    def test_isoformat01(self):
        self.assertEqual(dt3.isoformat("T"), "2002-03-01T12:59:59.000100+01:00")

    def test_isoformat02(self):
        self.assertEqual(dt3.isoformat(" "), "2002-03-01 12:59:59.000100+01:00")

    def test_isoformat03(self):
        self.assertEqual(str(dt3), "2002-03-01 12:59:59.000100+01:00")

    @unittest.skipIf(STDLIB, "standard datetime differs")
    def test___repr__00(self):
        self.assertEqual(repr(dt1), dt1r)

    @unittest.skipIf(STDLIB, "standard datetime differs")
    def test___repr__01(self):
        self.assertEqual(repr(dt3), dt3r)

    @unittest.skipIf(STDLIB, "standard datetime differs")
    def test___repr__02(self):
        self.assertEqual(repr(dt4), dt4r)

    def test___repr__03(self):
        self.assertEqual(dt1, eval_mod(repr(dt1)))

    def test___repr__04(self):
        self.assertEqual(dt3, eval_mod(repr(dt3)))

    def test___repr__05(self):
        self.assertEqual(dt4, eval_mod(repr(dt4)))

    def test___hash__00(self):
        self.assertEqual(dt1, dt5)
        self.assertEqual(hash(dt1), hash(dt5))

    def test___hash__01(self):
        dd1 = dt1 + timedelta(weeks=7)
        dd5 = dt5 + timedelta(days=7 * 7)
        self.assertEqual(hash(dd1), hash(dd5))

    def test___hash__02(self):
        d = {dt1: 1}
        d[dt5] = 2
        self.assertEqual(len(d), 1)
        self.assertEqual(d[dt1], 2)

    def test___hash__03(self):
        self.assertNotEqual(dt1, dt1z1)
        self.assertNotEqual(hash(dt1), hash(dt1z1))

    def test___hash__04(self):
        self.assertNotEqual(dt1z1, dt5z2)
        self.assertNotEqual(hash(dt1z1), hash(dt5z2))

    @unittest.skipIf(STDLIB, "standard datetime has no tuple()")
    def test_tuple00(self):
        self.assertEqual(dt1.tuple(), (2002, 1, 31, 0, 0, 0, 0, None, 0))

    @unittest.skipIf(STDLIB, "standard datetime has no tuple()")
    def test_tuple01(self):
        self.assertEqual(dt27tz2.tuple(), (2010, 3, 27, 12, 0, 0, 0, tz2, 0))

    @unittest.skipIf(STDLIB, "standard datetime has no tuple()")
    def test_tuple02(self):
        self.assertEqual(dt28tz2.tuple(), (2010, 3, 28, 12, 0, 0, 0, tz2, 0))


if __name__ == "__main__":
    unittest.main()



try:
    '' % ()
except TypeError:
    print("SKIP")
    raise SystemExit

print("%%" % ())
print("=%s=" % 1)
print("=%s=%s=" % (1, 2))
print("=%s=" % (1,))
print("=%s=" % [1, 2])

print("=%s=" % "str")
print("=%r=" % "str")

# test calling __int__
class A:
    def __int__(self):
        return 123
print("%d" % A())

try:
    print("=%s=%s=" % 1)
except TypeError:
    print("TypeError")

try:
    print("=%s=%s=%s=" % (1, 2))
except TypeError:
    print("TypeError")

try:
    print("=%s=" % (1, 2))
except TypeError:
    print("TypeError")

print("%s" % True)
print("%s" % 1)
print("%.1s" % "ab")

print("%r" % True)
print("%r" % 1)

print("%c" % 48)
print("%c" % 'a')
print("%10s" % 'abc')
print("%-10s" % 'abc')

print('%c' % False)
print('%c' % True)

# Should be able to print dicts; in this case they aren't used
# to lookup keywords in formats like %(foo)s
print('%s' % {})      # dict treated as the single (positional) arg to %
print('%s' % ({},))   # dict is the first (and only) arg in the positional arg tuple
print('foo' % {})     # no error, dict treated as an empty map of named args

# Cases when "*" used and there's not enough values total
try:
    print("%*s" % 5)
except TypeError:
    print("TypeError")
try:
    print("%*.*s" % (1, 15))
except TypeError:
    print("TypeError")

print("%(foo)s" % {"foo": "bar", "baz": False})
print("%s %(foo)s %(foo)s" % {"foo": 1})  # dict consumed positionally, then used as map - ok
try:
    print("%(foo)s %s %(foo)s" % {"foo": 1})  # used as map, then positionally - not enough args
except TypeError:
    print("TypeError")
try:
    print("%(foo)s" % {})
except KeyError:
    print("KeyError")
# Using in "*" with dict got to fail
try:
    print("%(foo)*s" % {"foo": "bar"})
except TypeError:
    print("TypeError")

# When using %(foo)s format the single argument must be a dict
try:
    '%(foo)s' % 1
except TypeError:
    print('TypeError')
try:
    '%(foo)s' % ({},)
except TypeError:
    print('TypeError')

try:
    '%(a' % {'a':1}
except ValueError:
    print('ValueError')

try:
    '%.*d %.*d' % (20, 5)
except TypeError:
    print('TypeError')

try:
    a = '%*' % 1
except (ValueError):
    print('ValueError')

try:
    '%c' % 'aa'
except TypeError:
    print('TypeError')

try:
    '%l' % 1
except ValueError:
    print('ValueError')

try:
    'a%' % 1
except ValueError:
    print('ValueError')



# Verify that sys is a builtin.
import sys

print(sys, hasattr(sys, "__file__"))

sys.path.clear()
sys.path.append("ext")

# All three should only get builtins, despite sys.py, usys.py, and
# micropython.py being in the path.
# usys isn't extensible, but has a special-cased alias for backwards
# compatibility.
import micropython

print(micropython, hasattr(micropython, "__file__"))
import sys

print(sys, hasattr(sys, "__file__"))
import usys

print(usys, hasattr(usys, "__file__"))

# This should get os.py, which uses uos to get the builtin.
import os

print(os, hasattr(os, "__file__"), os.sep, os.extra)

# This should get time.py, which uses empty sys.path to get the builtin.
import time

print(time, hasattr(time, "__file__"), time.sleep, time.extra)

# These should get the builtins.
import uos

print(uos, hasattr(uos, "__file__"), hasattr(uos, "extra"))

import utime

print(utime, hasattr(utime, "__file__"), hasattr(utime, "extra"))



import socket


CONTENT = b"""\
HTTP/1.0 200 OK

Hello #%d from MicroPython!
"""


def main(micropython_optimize=False):
    s = socket.socket()

    # Binding to all interfaces - server will be accessible to other hosts!
    ai = socket.getaddrinfo("0.0.0.0", 8080)
    print("Bind address info:", ai)
    addr = ai[0][-1]

    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.bind(addr)
    s.listen(5)
    print("Listening, connect your browser to http://<this_host>:8080/")

    counter = 0
    while True:
        res = s.accept()
        client_sock = res[0]
        client_addr = res[1]
        print("Client address:", client_addr)
        print("Client socket:", client_sock)

        if not micropython_optimize:
            # To read line-oriented protocol (like HTTP) from a socket (and
            # avoid short read problem), it must be wrapped in a stream (aka
            # file-like) object. That's how you do it in CPython:
            client_stream = client_sock.makefile("rwb")
        else:
            # .. but MicroPython socket objects support stream interface
            # directly, so calling .makefile() method is not required. If
            # you develop application which will run only on MicroPython,
            # especially on a resource-constrained embedded device, you
            # may take this shortcut to save resources.
            client_stream = client_sock

        print("Request:")
        req = client_stream.readline()
        print(req)
        while True:
            h = client_stream.readline()
            if h == b"" or h == b"\r\n":
                break
            print(h)
        client_stream.write(CONTENT % counter)

        client_stream.close()
        if not micropython_optimize:
            client_sock.close()
        counter += 1
        print()


main()



"""Filename globbing utility."""

import os
import os.path
import re
import fnmatch

__all__ = ["glob", "iglob"]


def glob(pathname):
    """Return a list of paths matching a pathname pattern.

    The pattern may contain simple shell-style wildcards a la
    fnmatch. However, unlike fnmatch, filenames starting with a
    dot are special cases that are not matched by '*' and '?'
    patterns.

    """
    return list(iglob(pathname))


def iglob(pathname):
    """Return an iterator which yields the paths matching a pathname pattern.

    The pattern may contain simple shell-style wildcards a la
    fnmatch. However, unlike fnmatch, filenames starting with a
    dot are special cases that are not matched by '*' and '?'
    patterns.

    """
    if not has_magic(pathname):
        if os.path.lexists(pathname):
            yield pathname
        return
    dirname, basename = os.path.split(pathname)
    if not dirname:
        for name in glob1(None, basename):
            yield name
        return
    # `os.path.split()` returns the argument itself as a dirname if it is a
    # drive or UNC path.  Prevent an infinite recursion if a drive or UNC path
    # contains magic characters (i.e. r'\\?\C:').
    if dirname != pathname and has_magic(dirname):
        dirs = iglob(dirname)
    else:
        dirs = [dirname]
    if has_magic(basename):
        glob_in_dir = glob1
    else:
        glob_in_dir = glob0
    for dirname in dirs:
        for name in glob_in_dir(dirname, basename):
            yield os.path.join(dirname, name)


# These 2 helper functions non-recursively glob inside a literal directory.
# They return a list of basenames. `glob1` accepts a pattern while `glob0`
# takes a literal basename (so it only has to check for its existence).


def glob1(dirname, pattern):
    if not dirname:
        if isinstance(pattern, bytes):
            dirname = bytes(os.curdir, "ASCII")
        else:
            dirname = os.curdir
    try:
        names = os.listdir(dirname)
    except os.error:
        return []
    if not _ishidden(pattern):
        names = [x for x in names if not _ishidden(x)]
    return fnmatch.filter(names, pattern)


def glob0(dirname, basename):
    if not basename:
        # `os.path.split()` returns an empty basename for paths ending with a
        # directory separator.  'q*x/' should match only directories.
        if os.path.isdir(dirname):
            return [basename]
    else:
        if os.path.lexists(os.path.join(dirname, basename)):
            return [basename]
    return []


magic_check = re.compile("[*?[]")
magic_check_bytes = re.compile(b"[*?[]")


def has_magic(s):
    if isinstance(s, bytes):
        match = magic_check_bytes.search(s)
    else:
        match = magic_check.search(s)
    return match is not None


def _ishidden(path):
    return path[0] in (".", b"."[0])



import sys
import os
import tarfile

if len(sys.argv) < 2:
    raise ValueError("Usage: %s inputfile.tar" % sys.argv[0])

t = tarfile.TarFile(sys.argv[1])
for i in t:
    print(i.name)
    if i.type == tarfile.DIRTYPE:
        os.mkdir(i.name)
    else:
        f = t.extractfile(i)
        with open(i.name, "wb") as of:
            of.write(f.read())



import sys

# ruff: noqa: E402
sys.path.insert(0, ".")
import aiohttp
import asyncio


params = {"key1": "value1", "key2": "value2"}


async def main():
    async with aiohttp.ClientSession() as session:
        async with session.get("http://httpbin.org/get", params=params) as response:
            expect = "http://httpbin.org/get?key1=value1&key2=value2"
            assert str(response.url) == expect, f"{response.url} != {expect}"
            html = await response.text()
            print(html)


asyncio.run(main())



# Example of using PIO for PWM, and fading the brightness of an LED

# ruff: noqa: F821 - @asm_pio decorator adds names to function scope

from machine import Pin
from rp2 import PIO, StateMachine, asm_pio
from time import sleep


@asm_pio(sideset_init=PIO.OUT_LOW)
def pwm_prog():
    # fmt: off
    pull(noblock) .side(0)
    mov(x, osr)  # Keep most recent pull data stashed in X, for recycling by noblock
    mov(y, isr)  # ISR must be preloaded with PWM count max
    label("pwmloop")
    jmp(x_not_y, "skip")
    nop()         .side(1)
    label("skip")
    jmp(y_dec, "pwmloop")
    # fmt: on


class PIOPWM:
    def __init__(self, sm_id, pin, max_count, count_freq):
        self._sm = StateMachine(sm_id, pwm_prog, freq=2 * count_freq, sideset_base=Pin(pin))
        # Use exec() to load max count into ISR
        self._sm.put(max_count)
        self._sm.exec("pull()")
        self._sm.exec("mov(isr, osr)")
        self._sm.active(1)
        self._max_count = max_count

    def set(self, value):
        # Minimum value is -1 (completely turn off), 0 actually still produces narrow pulse
        value = max(value, -1)
        value = min(value, self._max_count)
        self._sm.put(value)


# Pin 25 is LED on Pico boards
pwm = PIOPWM(0, 25, max_count=(1 << 16) - 1, count_freq=10_000_000)

while True:
    for i in range(256):
        pwm.set(i**2)
        sleep(0.01)



# Test executing Python code while IRQs disabled.
try:
    from machine import disable_irq, enable_irq
    from time import ticks_us
except ImportError:
    print("SKIP")
    raise SystemExit


# Structured to also nest disable_irq() calls
def f(n):
    st_irq = disable_irq()
    if n:
        f(n - 1)
    else:
        # busy-wait in a tight loop for 1ms, to simulate doing some work in a critical section
        # (can't wait for time_us() to increment as not all ports will "tick" with interrupts off.)
        for _ in range(100):
            ticks_us()
    enable_irq(st_irq)


for nest in range(3):
    print(nest)
    for _ in range(5):
        f(nest)



# Test asyncio TCP server and client with TLS, using readline() to read data.

try:
    import os
    import asyncio
    import ssl
except ImportError:
    print("SKIP")
    raise SystemExit

PORT = 8000

# These are test certificates. See tests/README.md for details.
cert = cafile = "ec_cert.der"
key = "ec_key.der"

try:
    os.stat(cafile)
    os.stat(key)
except OSError:
    print("SKIP")
    raise SystemExit


async def handle_connection(reader, writer):
    data = await reader.readline()
    print("echo:", data)
    data2 = await reader.readline()
    print("echo:", data2)
    writer.write(data + data2)
    await writer.drain()

    print("close")
    writer.close()
    await writer.wait_closed()

    print("done")
    ev.set()


async def tcp_server():
    global ev

    server_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    server_ctx.load_cert_chain(cert, key)
    ev = asyncio.Event()
    server = await asyncio.start_server(handle_connection, "0.0.0.0", PORT, ssl=server_ctx)
    print("server running")
    multitest.next()
    async with server:
        await asyncio.wait_for(ev.wait(), 10)


async def tcp_client(message):
    client_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    client_ctx.verify_mode = ssl.CERT_REQUIRED
    client_ctx.load_verify_locations(cafile=cafile)
    reader, writer = await asyncio.open_connection(
        IP, PORT, ssl=client_ctx, server_hostname="micropython.local"
    )
    print("write:", message)
    writer.write(message)
    await writer.drain()
    data = await reader.readline()
    print("read:", data)
    data2 = await reader.readline()
    print("read:", data2)


def instance0():
    multitest.globals(IP=multitest.get_network_ip())
    asyncio.run(tcp_server())


def instance1():
    multitest.next()
    asyncio.run(tcp_client(b"client data\nclient data2\n"))



# Test of a ESPnow echo server and asyncio client transferring data.
# Test will SKIP if instance1 (asyncio client) does not support asyncio.
# - eg. ESP8266 with 1MB flash.
# Instance0 is not required to support asyncio.

try:
    import network
    import random
    import espnow
except ImportError:
    print("SKIP")
    raise SystemExit

# Set read timeout to 5 seconds
timeout_ms = 5000
default_pmk = b"MicroPyth0nRules"
sync = True


def echo_server(e):
    peers = []
    while True:
        peer, msg = e.irecv(timeout_ms)
        if peer is None:
            return
        if peer not in peers:
            peers.append(peer)
            e.add_peer(peer)

        #  Echo the MAC and message back to the sender
        if not e.send(peer, msg, sync):
            print("ERROR: send() failed to", peer)
            return

        if msg == b"!done":
            return


def client_send(e, peer, msg, sync):
    print("TEST: send/recv(msglen=", len(msg), ",sync=", sync, "): ", end="", sep="")
    try:
        if not e.send(peer, msg, sync):
            print("ERROR: Send failed.")
            return
    except OSError as exc:
        # Don't print exc as it is differs for esp32 and esp8266
        print("ERROR: OSError:")
        return
    print("OK")


def init(e, sta_active=True, ap_active=False):
    wlans = [network.WLAN(i) for i in [network.STA_IF, network.AP_IF]]
    e.active(True)
    e.set_pmk(default_pmk)
    wlans[0].active(sta_active)
    wlans[1].active(ap_active)
    wlans[0].disconnect()  # Force esp8266 STA interface to disconnect from AP
    return e


async def client(e):
    init(e, True, False)
    e.config(timeout_ms=timeout_ms)
    peer = PEERS[0]
    e.add_peer(peer)
    multitest.next()

    print("airecv() test...")
    msgs = []
    for i in range(5):
        # Send messages to the peer who will echo it back
        msgs.append(bytes([random.getrandbits(8) for _ in range(12)]))
        client_send(e, peer, msgs[i], True)

    for i in range(5):
        mac, reply = await e.airecv()
        print("OK" if reply == msgs[i] else "ERROR: Received != Sent")

    # Tell the server to stop
    print("DONE")
    msg = b"!done"
    client_send(e, peer, msg, True)
    mac, reply = await e.airecv()
    print("OK" if reply == msg else "ERROR: Received != Sent")

    e.active(False)


# Server
def instance0():
    e = espnow.ESPNow()
    init(e, True, False)
    multitest.globals(PEERS=[network.WLAN(i).config("mac") for i in (0, 1)])
    multitest.next()
    print("Server Start")
    echo_server(e)
    print("Server Done")
    e.active(False)


# Client
def instance1():
    try:
        import asyncio
        from aioespnow import AIOESPNow
    except ImportError:
        print("SKIP")
        raise SystemExit
    asyncio.run(client(AIOESPNow()))



# test calling a function with multiple **args


def f(a, b=None, c=None):
    print(a, b, c)


f(**{"a": 1}, **{"b": 2})
f(**{"a": 1}, **{"b": 2}, c=3)
f(**{"a": 1}, b=2, **{"c": 3})

try:
    f(1, **{"b": 2}, **{"b": 3})
except TypeError:
    print("TypeError")

# test calling a method with multiple **args


class A:
    def f(self, a, b=None, c=None):
        print(a, b, c)


a = A()
a.f(**{"a": 1}, **{"b": 2})
a.f(**{"a": 1}, **{"b": 2}, c=3)
a.f(**{"a": 1}, b=2, **{"c": 3})

try:
    a.f(1, **{"b": 2}, **{"b": 3})
except TypeError:
    print("TypeError")



"""
categories: Syntax
description: Argument unpacking does not work if the argument being unpacked is the nth or greater argument where n is the number of bits in an MP_SMALL_INT.
cause: The implementation uses an MP_SMALL_INT to flag args that need to be unpacked.
workaround: Use fewer arguments.
"""


def example(*args):
    print(len(args))


MORE = ["a", "b", "c"]

# fmt: off
example(
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
    16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
    32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
    48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63,
    *MORE,
)
# fmt: on



# Test a UTF-8 encoded literal
s = "asdf©qwer"
for i in range(len(s)):
    print("s[%d]: %s   %X" % (i, s[i], ord(s[i])))

# Test all three forms of Unicode escape, and
# all blocks of UTF-8 byte patterns
s = "a\xA9\xFF\u0123\u0800\uFFEE\U0001F44C"
for i in range(-len(s), len(s)):
    print("s[%d]: %s   %X" % (i, s[i], ord(s[i])))
    print("s[:%d]: %d chars, '%s'" % (i, len(s[:i]), s[:i]))
    for j in range(i, len(s)):
        print("s[%d:%d]: %d chars, '%s'" % (i, j, len(s[i:j]), s[i:j]))
    print("s[%d:]: %d chars, '%s'" % (i, len(s[i:]), s[i:]))

# Test UTF-8 encode and decode
enc = s.encode()
print(enc, enc.decode() == s)

# printing of unicode chars using repr
# NOTE: for some characters (eg \u10ff) we differ to CPython
print(repr("a\uffff"))
print(repr("a\U0001ffff"))

# test invalid escape code
try:
    eval('"\\U00110000"')
except SyntaxError:
    print("SyntaxError")

# test unicode string given to int
try:
    int("\u0200")
except ValueError:
    print("ValueError")

# test invalid UTF-8 string
try:
    str(b"ab\xa1", "utf8")
except UnicodeError:
    print("UnicodeError")
try:
    str(b"ab\xf8", "utf8")
except UnicodeError:
    print("UnicodeError")
try:
    str(bytearray(b"ab\xc0a"), "utf8")
except UnicodeError:
    print("UnicodeError")
try:
    str(b"\xf0\xe0\xed\xe8", "utf8")
except UnicodeError:
    print("UnicodeError")



"""
The MIT License (MIT)

Copyright (c) 2023 Arduino SA
Copyright (c) 2018 KPN (Jan Bogaerts)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from senml import *


def do_actuate(record):
    """
    called when actuate_me receives a value.
    :return: None
    """
    print("for known device: ")
    print(record.value)


def device_callback(record, **kwargs):
    """
    a generic callback, attached to the device. Called when a record is found that has not yet been registered
    in the pack. When this callback is called, the record will already be added to the pack.
    :param kwargs: optional extra parameters
    :param record: the newly found record.
    :return: None
    """
    print("found record: " + record.name)
    print("with value: " + record.value)


def gateway_callback(record, **kwargs):
    """
    a generic callback, attached to the device. Called when a record is found that has not yet been registered
    in the pack. When this callback is called, the record will already be added to the pack.
    :param record: the newly found record.
    :param kwargs: optional extra parameters (device can be found here)
    :return: None
    """
    if "device" in kwargs and kwargs["device"] is not None:
        print("for device: " + kwargs["device"].name)
    else:
        print("for gateway: ")
    print("found record: " + record.name)
    print("with value: " + str(record.value))


gateway = SenmlPack("gateway_name", gateway_callback)
device = SenmlPack("device_name", device_callback)
actuate_me = SenmlRecord("actuator", callback=do_actuate)

gateway.add(device)
device.add(actuate_me)
gateway.from_json(
    '[{"bn": "gateway_name", "n":"temp", "v": 22},{"n": "gateway_actuator", "vb": true}, {"bn": "device_name", "n":"actuator", "v": 20 }, {"n": "another_actuator", "vs": "a value"}, {"bn": "device_2", "n":"temp", "v": 20 }, {"n": "actuator2", "vs": "value2"}]'
)
